#!/usr/bin/env python
"""This module contains utility functions for plotting:

    :func:`get_fig_axes`
        Helper function to check arguments of plotting functions. Retrieve figure
        and axes from `axes`. If `axes` is None, create figure and axes, and
        return those.

    :func:`split_axes`
        Split a :class:`matplotlib.axes.Axes` into one or more panels, with tied
        x and y axes. Also hides overlapping tick labels

    :func:`clean_invalid`
        Remove pairs of values from two arrays of data if either  element in the
        pair is `nan` or `inf`. Used to prepare data for histogram or violin
        plots; as even masked `nan` and `inf` values are not handled by these
        functions.

    :func:`get_kde`
        Evaluate a kernel density estimate over observed data in linear or
        log-transformed space (e.g. for making violin plots in log space,
        but having kernels appropriately scaled).
"""
import numpy
import scipy.stats
import matplotlib
import matplotlib.pyplot as plt


def get_fig_axes(axes=None):
    """Retrieve figure and axes from `axes`. If `axes` is None, both.
    
    Used as a helper function for replotting atop existing axes, by functions
    defined in :mod:`plastid.plotting.plots`.


    Parameters
    ----------
    axes : :class:`matplotlib.axes.Axes` or `None`
        Axes in which to place plot. If `None`, a new figure is generated.


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    :class:`matplotlib.axes.Axes`
        Axes containing plot
    """
    if axes is None:
        fig = plt.figure()
        ax = plt.gca()
    else:
        ax = axes
        fig = ax.figure

    return fig, ax

def split_axes(
        ax,
        top_height      = 0,
        left_width      = 0,
        right_width     = 0,
        bottom_height   = 0,
        main_ax_kwargs  = {},
        other_ax_kwargs = {}
): # yapf: disable
    """Split the spaces taken by one axes into one or more panes, setting the original axes invisible.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axes to split

    top_height, left_width, right_width, bottom_height : float, optional
        If not `None`, a panel on the corresponding side of the `ax` will
        be created, using whatever fraction is specified (e.g. 0.1 to use
        10% of total height).

    main_ax_kwargs : dict
        Dictionary of keyword arguments for central panes, passed
        to :meth:`matplotlib.figure.Figure.add_axes`

    other_ax_kwargs : dict
        Dictionary of keyword arguments for peripheral panes, passed
        to :meth:`matplotlib.figure.Figure.add_axes`


    Returns
    -------
    dict
        Dictionary of axes. `'orig'` refers to `ax`. The central panel is `'main'`.
        Other panels will be mapped to `'top'`, `'left`' et c, if they are created.
    """
    fig = ax.figure
    ax.set_visible(False)
    axes = {"orig": ax}
    mplrc = matplotlib.rcParams

    # yapf: disable
    buf_left  = mplrc["figure.subplot.left"]
    buf_bot   = mplrc["figure.subplot.bottom"]
    buf_right = 1.0 - mplrc["figure.subplot.right"]
    buf_top   = 1.0 - mplrc["figure.subplot.top"]

    hscale = 1.0 - buf_top - buf_bot
    wscale = 1.0 - buf_left - buf_right

    main_height = (1.0 - bottom_height - top_height) * hscale
    main_width = (1.0 - left_width - right_width) * wscale

    bottom_height *= hscale
    top_height    *= hscale
    left_width    *= wscale
    right_width   *= wscale

    main_left  = buf_left  + left_width
    main_right = main_left + main_width
    main_bot   = buf_bot   + bottom_height
    main_top   = main_bot  + main_height
    # yapf: enable

    # each rect is (left, bottom, width, height)
    rects = {}
    rects["main"] = [main_left, main_bot, main_width, main_height]
    if left_width > 0:
        rects["left"] = [buf_left, main_bot, left_width, main_height]
    if right_width > 0:
        rects["right"] = [main_right, main_bot, right_width, main_height]
    if bottom_height > 0:
        rects["bottom"] = [main_left, main_bot, main_width, bottom_height]
    if top_height > 0:
        rects["top"] = [main_left, main_top, main_width, top_height]

    axes["main"] = fig.add_axes(rects["main"], zorder=100, **main_ax_kwargs)

    for axes_name, rect in rects.items():
        if axes_name == "main":
            pass
        else:
            ax_kwargs = other_ax_kwargs
            if axes_name in ("right", "left"):
                ax_kwargs["sharey"] = axes["main"]
                if "sharex" in ax_kwargs:
                    ax_kwargs.pop("sharex")
            if axes_name in ("top", "bottom"):
                ax_kwargs["sharex"] = axes["main"]
                if "sharey" in ax_kwargs:
                    ax_kwargs.pop("sharey")

            axes[axes_name] = fig.add_axes(rect, zorder=50, **ax_kwargs)

    if "top" in axes:
        axes["top"].xaxis.tick_top()
        axes["main"].xaxis.tick_bottom()
        # prevent tick collisions
        axes["top"].yaxis.get_ticklabels()[0].set_visible(False)

    if "bottom" in axes:
        axes["bottom"].xaxis.tick_bottom()
        for t in axes["main"].xaxis.get_ticklabels():
            t.set_visible(False)

    if "left" in axes:
        axes["left"].yaxis.tick_left()
        axes["left"].xaxis.get_ticklabels()[-1].set_visible(False)
        for t in axes["main"].yaxis.get_ticklabels():
            t.set_visible(False)

    if "right" in axes:
        axes["right"].yaxis.tick_right()
        axes["main"].yaxis.tick_left()
        axes["right"].xaxis.get_ticklabels()[0].set_visible(False)

    return axes


def clean_invalid(x, y, min_x=-numpy.inf, min_y=-numpy.inf, max_x=numpy.inf, max_y=numpy.inf):
    """Remove corresponding values from x and y when one or both of those is `nan` or `inf`,
    and optionally truncate values to minima and maxima

    Parameters
    ----------
    x, y : :class:`numpy.ndarray` or list
        Pair arrays or lists of corresponding numbers

    min_x, min_y, max_x, max_y : number, optional
        If supplied, set values below `min_x` to `min_x`, values larger
        than `max_x` to `max_x` and so for `min_y` and `max_y`

    Returns
    -------
    :class:`numpy.ndarray`
        A shortened version of `x`, excluding invalid values

    :class:`numpy.ndarray`
        A shortened version of `y`, excluding invalid values
    """
    x = numpy.array(x).astype(float)
    y = numpy.array(y).astype(float)

    x[x < min_x] = min_x
    x[x > max_x] = max_x
    y[y < min_y] = min_y
    y[y > max_y] = max_y

    newmask = numpy.isinf(x) | numpy.isnan(x) | numpy.isinf(y) | numpy.isnan(y)
    x = x[~newmask]
    y = y[~newmask]

    return x, y


def get_kde(data, log=False, base=2, points=100, bw_method="scott"):
    """Estimate a kernel density (kde) over `data`

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Data to build kde over

    log : bool, optional
        If `True`, `data` is log-transformed before the kde is estimated.
        Data are converted back to non-log space afterwards.

    base : 2, 10, or :obj:`numpy.e`, optional
        If `log` is `True`, this serves as the base of the log space.
        If `log` is `False`, this is ignored. (Default: 2)

    points : int
        Number of points over which to evaluate kde. (Default: 100)

    bw_method : str
        Bandwith estimation method. See documentation for
        :obj:`scipy.stats.gaussian_kde`. (Default: "scott")

    Returns
    -------
    :class:`numpy.ndarray`
        Points over which kde is evaluated (x-values), in non-log space

    :class:`numpy.ndarray`
        Value of kde (y-values), in non-log space
    """
    if log == True:
        if base == 2:
            func = numpy.log2
        elif base == 10:
            func = numpy.log10
        elif base == numpy.e:
            func = numpy.log
        else:
            raise ValueError("kde: Base must be 2, 10, or numpy.e")

        data = func(data)
        domain = func(numpy.logspace(data.min(), data.max(), base=base, num=points))
    else:
        domain = numpy.linspace(data.min(), data.max(), points)

    kde = scipy.stats.gaussian_kde(data, bw_method=bw_method)

    curve = kde.evaluate(domain)

    if log == True:
        domain = base**domain

    return domain, curve



#!/usr/bin/env python
"""This module contains post-processors that reformats module docstrings for
use as command-line script help, by removing rich formatting, substitutions,
`reStructuredText`_ markup, and some `numpydoc`_ tokens.

See also
--------
:mod:`re`
    Python regular expressions module

`reStructuredText <http://docutils.sourceforge.net/rst.html>`_
    Specification for `reStructuredText`_, a markup language used by the
    `Sphinx`_ documentation engine, and used throughout the docstrings found
    in this package

`numpydoc <https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt>`_
    A docstring standard and `Sphinx`_ extension that converts human-readable
    docstrings into valid `reStructuredText`_ 
"""
import re


def shorten_help(inp):
    """Pretty prints `numpydoc`_-formatted module, class, or function docstrings
    for use in command-line help, by removing some `reStructuredText`_ markup,
    and truncating the docstring at the appropriate `numpydoc`_ tokens.
    
    Parameters
    ----------
    inp : multi-line str
        Class or function docstring to format
    
    Returns
    -------
    str
        Cleaned helptext
    """
    inp = pyrst_pattern.sub(r"\g<spacing>\g<argument>", inp)
    inp = subst_pattern.sub(r"\g<1>", inp)
    inp = link_pattern.sub(r"\g<1>", inp)

    tokens = ["Parameters", "Returns", "Yields", "Raises", "Attributes"]

    indices = [inp.find("%s" % X) for X in tokens]
    indices.extend([inp.find("    %s" % X.lower()) for X in tokens])

    for n, idx in enumerate(indices):
        if idx == -1:
            indices[n] = len(inp)

    return inp[:min(indices)].strip() + "\n"


def format_module_docstring(inp):
    """Pretty prints module docstrings for use in command-line help,
    by removing some `reStructuredText`_ markup, truncating the docstring
    at the first `numpydoc`_ token, and surrounding help text with separators.
    
    Parameters
    ----------
    inp : multi-line str
        Module docstring to format
    
    Returns
    -------
    str
        Formatted docstring
    """
    return _separator + "\n" + shorten_help(inp) + "\n" + _separator


pyrst_pattern = re.compile(
    r"(?P<spacing>^|\s+)(?::(?P<domain>[^:`<>]+))?:(?P<role>[^:`]*):`(?P<argument>[^`<>]+)(?: +<(?P<pointer>[^`]+)>)?`"
)
"""RegEx pattern that detects `reStructuredText`_ markup of python tokens
of the form ``:domain:role:`argument``` or simply ``:role:`argument```,
if the token is preceded by whitespace or begins a line.

Retrieves domain, role, and argument as named attributes if
:py:meth:`pyrst_pattern.groupdict` is called.
"""

subst_pattern = re.compile(r"\|([^|]*)\|")
"""RegEx pattern that matches `reStructuredText`_ substitution tokens
of form ``|substitution|``
"""

link_pattern = re.compile(r"`([^`<>]+)( <[^`]+>)?`_")
"""RegEx pattern that matches `reStructuredText`_ link references of forms ```Linkname`_``
and ```Link text <url>`_``
"""

_separator = "\n" + (78 * "-") + "\n"
"""78-dash long text separator for separating help sections in command-line environments"""



#!/usr/bin/env python
"""Convert transcripts from `BED`_, `BigBed`_, `GTF2`_, `GFF3`_, or `PSL`_ format
to `BED`_, :term:`extended BED`, or `GTF2`_ format.

 .. note::

    GFF3 schemas vary
        Different GFF3s have different schemas of hierarchy. By default, we
        assume the ontology used by the Sequence Ontology consortium. Users
        that require a different schema may supply `transcript_types` and
        `exon_types`, to indicate which sorts of features should be included.

    Identity relationships between elements vary between GFF3 files
        GFF3 files can represent discontiguous features using two strategies. In 
        one strategy, the exons of a transcript have unique IDs, but will share
        contain the same parent ID in their same `Parent` attribute in column 9
        of the GFF. In another strategy different exons of the same transcript
        simply share the same ID, and don't define a `Parent`. Here, both schemes
        are accepted, although what happens if they conflict within a single
        transcript is undefined.

"""
import argparse
import inspect
import warnings
import sys
import os
import copy

from plastid.util.scriptlib.argparsers import AnnotationParser, BaseParser
from plastid.util.services.exceptions import ArgumentWarning, warn
from plastid.util.io.openers import argsopener, get_short_name
from plastid.util.io.filters import NameDateWriter
from plastid.util.scriptlib.help_formatters import format_module_docstring

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

MAKE_BIGBED_MESSAGE =\
"""The BED-formatted file `%(outbase)s` has been exported with extended BED
columns. Plastid and other software can detect these automatically if you convert
`%(outbase)s` to a BigBed file using Jim Kent's bedToBigBed utility, as follows:

    $ sort -k1,1 -k2,2n -t"\t" %(outbase)s >%(outbase)s_sorted.bed
    $ bedToBigBed -tab -type=bed12+%(numcols)s -as=fields.as %(outbase)s.bed -extraIndex=name chrom.sizes %(outbase)s.bb

where `chrom.sizes` is a two-column, tab-delimited table of chromosome name & size,
and `fields.as` is an autoSql declaration describing the types & names of the extra
columns. For `%(outbase)s.bed`, `fields.as` could be:

%(autosql)s

Simply edit the type of each custom field to match your data type. Field type
options are:

    int, uint, short, ushort, byte, ubyte, float, char, string, lstring, enum, set

For details, See the autoSql Grammar specification at:

https://github.com/ENCODE-DCC/kentUtils/blob/36d6274459f644d5400843b8fa097b380b8f7867/src/hg/autoSql/autoSql.doc
"""

BED12_RESERVED_NAMES = [
    "chrom",
    "chromStart",
    "chromEnd",
    "name",
    "score",
    "strand",
    "thickStart",
    "thickEnd",
    "reserved",
    "blockCount",
    "blockSizes",
    "chromStarts",
]

DEFAULT_AUTOSQL_STR =\
"""
table bigbed_columns "%s columns"
(
    string            chrom;          "Chromosome"
    uint              chromStart;     "chr start"
    uint              chromEnd;       "chr end"
    string            name;           "item name"
    uint              score;          "score"
    char[1]           strand;         "strand"
    uint              thickStart;     "thickstart"
    uint              thickEnd;       "thickend"
    uint              reserved;       "normally itemRgb"
    int               blockCount;     "block count"
    int[blockCount]   blockSizes;     "block sizes"
    int[blockCount]   chromStarts;    "block starts"
%s
)
"""

AUTOSQL_ROW_FMT_STR = '    string            %s;%s"description of custom field contents"'


def fix_name(inp, names_used):
    """Append a number if an autoSql field name is duplicated.
    """
    name = inp
    i = 2
    while name in names_used:
        name = "%s%s" % (inp, i)
        i += 1

    names_used.append(name)
    return name


#: TODO: no functional test for --extra_columns
def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: sys.argv[1:] (actually command-line arguments)
    """
    ap = AnnotationParser()
    bp = BaseParser()
    annotation_parser = ap.get_parser()
    base_parser = bp.get_parser()

    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        parents=[base_parser, annotation_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "--no_escape",
        default=True,
        action="store_false",
        help=
        "If specified and output format is GTF2, special characters in column 9 will be escaped (default: True)"
    )
    parser.add_argument(
        "--output_format",
        choices=["BED", "GTF2"],
        default="GTF2",
        help="Format of output file. (default: GTF2)"
    )
    parser.add_argument(
        "--extra_columns",
        nargs="+",
        default=[],
        type=str,
        help=
        "Attributes (e.g. 'gene_id' to output as extra columns in extended BED format (BED output only)."
    )
    parser.add_argument(
        "--empty_value",
        default="na",
        type=str,
        help=
        "Value to use of an attribute in `extra_columns` is not defined for a particular record (Default: 'na'"
    )
    parser.add_argument("outfile", metavar="outfile.[ bed | gtf ]", type=str, help="Output file")
    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    end_message = ""
    extra_cols = args.extra_columns
    if extra_cols is not None:
        if args.output_format == "BED":

            # avoid name clashes
            names_used = copy.copy(BED12_RESERVED_NAMES)
            asql_names = [fix_name(X, names_used) for X in extra_cols]
            autosql_str = "\n".join(
                AUTOSQL_ROW_FMT_STR % (X, " " * max(15 - len(X), 2)) for X in asql_names
            )

            file_info = {
                "outbase": args.outfile.replace(".bed", "").replace(".gtf", ""),
                "numcols": len(extra_cols),
                "autosql": DEFAULT_AUTOSQL_STR % (os.path.basename(args.outfile[:-4]), autosql_str),
            }
            end_message = MAKE_BIGBED_MESSAGE % file_info
        else:
            warn(
                "`--extra_columns` is ignored for %s-formatted output." % (args.output_format),
                ArgumentWarning
            )

    with argsopener(args.outfile, args, "w") as fout:
        c = 0
        transcripts = ap.get_transcripts_from_args(args, printer=printer)

        for transcript in transcripts:
            if args.output_format == "GTF2":
                fout.write(transcript.as_gtf(escape=args.no_escape))
            elif args.output_format == "BED":
                fout.write(
                    transcript.as_bed(extra_columns=extra_cols, empty_value=args.empty_value)
                )
            if c % 1000 == 1:
                printer.write("Processed %s transcripts ..." % c)
            c += 1

    printer.write("Processed %s transcripts total." % c)
    printer.write("Done.")
    print(end_message)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This module contains |BED_Reader|, an iterator that reads each line of a `BED`_
or :term:`extended BED` file into a |SegmentChain|, |Transcript|, or similar object. 

.. contents::
   :local:

Module contents
---------------

.. autosummary::

   BED_Reader
   bed_x_formats

Examples
--------
Read entries in a `BED`_ file as |Transcripts|. `thickEnd` and `thickStart`
columns will be interpreted as the endpoints of coding regions::

    >>> bed_reader = BED_Reader("some_file.bed",return_type=Transcript)
    >>> for transcript in bed_reader:
            pass # do something fun with each Transcript/SegmentChain

If `return_type` is unspecified, `BED`_ lines are read as |SegmentChains|::

    >>> my_chains = list(BED_Reader("some_file.bed"))
    >>> my_chains[:5]
        [list of segment chains as output...]

Open an :term:`extended BED` file, which contains additional columns for `gene_id`
and `favorite_color`. Values for these attributes will be stored in the `attr`
dict of each |Transcript|::

    >>> bed_reader = BED_Reader("some_file.bed",return_type=Transcript,extra_columns=["gene_id","favorite_color"])

Open several `Tabix`_-compressed `BED`_ files, and iterate over them as if
they were one stream::

    >>> bed_reader = BED_Reader("file1.bed.gz","file2.bed.gz",tabix=True)
    >>> for chain in bed_reader:
    >>>     pass # do something interesting with each chain
                                

See Also
--------
`UCSC file format FAQ <http://genome.ucsc.edu/FAQ/FAQformat.html>`_.
    BED format specification at UCSC
"""
__date__ = "Aug 23, 2011"
__author__ = "joshua"

import shlex
from plastid.readers.common import AssembledFeatureReader
from plastid.util.services.exceptions import FileFormatWarning, warn

bed_x_formats = {
    "bedDetail": [
        ("ID", str),
        ("description", str),
    ],
    "narrowPeak": [("signalValue", float),
                   ("pValue", float),
                   ("qValue", float),
                   ("peak", int)],
    "broadPeak": [("signalValue", float),
                  ("pValue", float),
                  ("qValue", float)],
    "gappedPeak": [("signalValue", float),
                   ("pValue", float),
                   ("qValue", float)],
    "tagAlign": [("sequence", str),
                 ("score", float),
                 ("strand", str)],
    "pairedTagAlign": [("seq1", str),
                       ("seq2", str)],
    "peptideMapping":
    [("rawScore", float),
     ("spectrumId", str),
     ("peptideRank", int),
     ("peptideRepeatCount", int)],
}
"""Column names and types for various :term:`extended BED` formats used by 
the `ENCODE`_ project. These can be passed to the `extra_columns` keyword of
:class:`BED_Reader`."""


class BED_Reader(AssembledFeatureReader):
    """
    BED_Reader(*streams, return_type=SegmentChain, add_three_for_stop=False, extra_columns=0, printer=None, tabix=False)
    
    Reads `BED`_ and :term:`extended BED` files line-by-line into |SegmentChains|
    or |Transcripts|. Metadata, if present in a track declaration, is saved
    in `self.metadata`. Malformed lines are stored in `self.rejected`, while
    parsing continues.

    Parameters
    ----------
    *streams : file-like
        One or more open filehandles of input data.
    
    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If set to
        `True`, three nucleotides will be added to the threeprime end of each
        CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
        feature. (Default: `False`)

    extra_columns: int or list optional
        Extra, non-BED columns in :term:`extended BED` format file corresponding
        to feature attributes. This is common in `ENCODE`_-specific `BED`_ variants.
        
        if `extra-columns` is:
        
          - an :class:`int`: it is taken to be the
            number of attribute columns. Attributes will be stored in
            the `attr` dictionary of the |SegmentChain|, under names like
            `custom0`, `custom1`, ... , `customN`.

          - a :class:`list` of :class:`str`, it is taken to be the names
            of the attribute columns, in order, from left to right in the file.
            In this case, attributes in extra columns will be stored under
            their respective names in the `attr` dict.

          - a :class:`list` of :class:`tuple`, each tuple is taken
            to be a pair of `(attribute_name, formatter_func)`. In this case,
            the value of `attribute_name` in the `attr` dict of the |SegmentChain|
            will be set to `formatter_func(column_value)`.
        
        (Default: 0)
                    
    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|
    
    tabix : boolean, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)
        

    Examples
    --------
    Read entries in a `BED`_ file as |Transcripts|. `thickEnd` and `thickStart`
    columns will be interpreted as the endpoints of coding regions::
    
        >>> bed_reader = BED_Reader(open("some_file.bed"),return_type=Transcript)
        >>> for transcript in bed_reader:
        >>>     pass # do something fun
    
    Open an :term:`extended BED` file that contains additional columns for `gene_id`
    and `favorite_color`. Values for these attributes will be stored in the `attr`
    dict of each |Transcript|::
    
        >>> bed_reader = BED_Reader(open("some_file.bed"),return_type=Transcript,extra_columns=["gene_id","favorite_color"])

    Open several `Tabix`_-compressed `BED`_ files, and iterate over them as if
    they were one uncompressed stream::
    
        >>> bed_reader = BED_Reader("file1.bed.gz","file2.bed.gz",tabix=True)
        >>> for chain in bed_reader:
        >>>     pass # do something more interesting

    
    Attributes
    ----------
    streams : file-like
        One or more open streams (usually filehandles) of input data.
    
    return_type : class
        The type of object assembled by the reader. Typically a |SegmentChain|
        or a subclass thereof. Must import a method called ``from_bed()``

    counter : int
        Cumulative line number counter over all streams
    
    rejected : list
        List of `BED`_ lines that could not be parsed
    
    metadata : dict
        Attributes declared in track line, if any
    
    extra_columns : int or list, optional
        Extra, non-BED columns in :term:`extended BED` format file corresponding to feature
        attributes. This is common in `ENCODE`_-specific `BED`_ variants.
        
        if `extra_columns` is:
        
          - an :class:`int`: it is taken to be the
            number of attribute columns. Attributes will be stored in
            the `attr` dictionary of the |SegmentChain|, under names like
            `custom0`, `custom1`, ... , `customN`.

          - a :class:`list` of :class:`str`, it is taken to be the names
            of the attribute columns, in order, from left to right in the file.
            In this case, attributes in extra columns will be stored under
            there respective names in the `attr` dict.

          - a :class:`list` of :class:`tuple`, each tuple is taken
            to be a pair of `(attribute_name, formatter_func)`. In this case,
            the value of `attribute_name` in the `attr` dict of the |SegmentChain|
            will be set to `formatter_func(column_value)`.
        
        If unspecified, :class:`BED_Reader` reads the track declaration line
        (if present), and:

          - if a known track type is specified by the `type` field, it attempts
            to format the extra columns as specified by that type. Known track
            types presently include:

              - bedDetail
              - narrowPeak
              - broadPeak
              - gappedPeak
              - tagAlign
              - pairedTagAlign
              - peptideMapping
          
          - if not, it assumes 0 non-`BED`_ fields are present, and that all columns
            are `BED`_ formatted.
    """

    def __init__(self, *args, **kwargs):
        """
        BED_Reader(*streams, return_type=SegmentChain, add_three_for_stop=False, extra_columns=0, printer=None, tabix=False)
        
        Parameters
        ----------
        *streams : file-like
            One or more open filehandles of input data.
        
        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations. If set to
            `True`, three nucleotides will be added to the threeprime end of each
            CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
            feature. (Default: `False`)

        extra_columns: int or list optional
            Extra, non-BED columns in :term:`Extended BED`_ format file corresponding
            to feature attributes. This is common in `ENCODE`_-specific `BED`_ variants.
            
            if `extra-columns` is:
            
              - an :class:`int`: it is taken to be the
                number of attribute columns. Attributes will be stored in
                the `attr` dictionary of the |SegmentChain|, under names like
                `custom0`, `custom1`, ... , `customN`.

              - a :class:`list` of :class:`str`, it is taken to be the names
                of the attribute columns, in order, from left to right in the file.
                In this case, attributes in extra columns will be stored under
                their respective names in the `attr` dict.

              - a :class:`list` of :class:`tuple`, each tuple is taken
                to be a pair of `(attribute_name, formatter_func)`. In this case,
                the value of `attribute_name` in the `attr` dict of the |SegmentChain|
                will be set to `formatter_func(column_value)`.
            
            (Default: 0)
                        
        printer : file-like, optional
            Logger implementing a ``write()`` method. Default: |NullWriter|
        
        tabix : boolean, optional
            `streams` are `tabix`_-compressed (Default: `False`)
        """
        AssembledFeatureReader.__init__(self, *args, **kwargs)
        self.extra_columns = kwargs.get("extra_columns", 0)

    def _parse_track_line(self, inp):
        """Parse track line from `BED`_ / extended BED file
        
        Parameters
        ----------
        inp : str
            track definition line from `BED`_  / extended BED file

        Returns
        -------
        dict
            key-value pairs from `BED`_ line
        """
        self.metadata = {}
        ltmp = shlex.split(inp.strip("\n"))
        for item in ltmp:
            k, v = item.split("=")
            self.metadata[k] = v

        track_type = self.metadata.get("type", None)
        if track_type is not None:
            if track_type in bed_x_formats:
                self.printer.write(
                    "Found track type '%s' in track definition line. Assuming extra columns follow UCSC definitions."
                    % track_type
                )
                if self.extra_columns == 0:
                    self.extra_columns = bed_x_formats[track_type]
                elif self.extra_columns != bed_x_formats[track_type]:
                    my_columns = self._get_extra_column_names()
                    track_format_columns = ",".join([X[0] for X in bed_x_formats[track_type]])
                    warn("Extra columns specified by %s track type declaration (%s) don't match those specified by user (%s). Using those specified by user." %\
                         (track_type,track_format_columns,my_columns),FileFormatWarning)
                    self.metadata["type"] = "custom"
            else:
                self.printer.write("Found track type '%s' in track definition line." % track_type)

    def _get_extra_column_names(self):
        """Return names of extra columns in extended BED file)"""
        if isinstance(self.extra_columns, int):
            my_columns = "%s unnamed columns" % self.extra_columns
        elif isinstance(self.extra_columns, list):
            if all([isinstance(X, tuple) for X in self.extra_columns]):
                my_columns = ",".join([X[0] for X in self.extra_columns])
            elif all([isinstance(X, str) for X in self.extra_columns]):
                my_columns = ",".join(self.extra_columns)

        return my_columns

    def _assemble(self, line):
        """Read `BED`_ files line-by-line into types specified by `self.return_type`"""
        self.counter += 1
        if line.strip() == "":
            return self.__next__()
        elif line.startswith("browser"):
            return self.__next__()
        elif line.startswith("track"):
            # reset metadata
            self._parse_track_line(line[5:])
            return self.__next__()
        elif line.startswith("#"):
            return self.__next__()
        else:
            try:
                return self.return_type.from_bed(line, extra_columns=self.extra_columns)
            except:
                self.rejected.append(line)
                msg = "Cannot parse BED line number %s. " % self.counter
                if self.metadata.get("type", None) is not None:
                    msg += (
                        "Are you sure this is a %s BED file with extra columns (%s)?" %
                        (self.metadata.get("type"), self._get_extra_column_names())
                    )
                elif self.extra_columns != 0:
                    msg += (
                        "Are you sure this BED file has extra columns (%s)?" %
                        self._get_extra_column_names()
                    )
                else:
                    msg += "Maybe this BED has extra columns (i.e. is an extended BED file)?"

                msg += ("\n    %s" % line)
                warn(msg, FileFormatWarning)
                return self.__next__()



#!/usr/bin/env python
"""Utilities for handling and wrapping various I/O operations.

Package overview
================

    =======================================  ======================================
    **Package module**                       **Contents**
    ---------------------------------------  --------------------------------------
    :py:mod:`~plastid.util.io.binary`        Tools for unpacking binary values
                                             into named dictionaries 

    :py:mod:`~plastid.util.io.filters`       Wrappers that transform input or
                                             output during file I/O (e.g. remove
                                             comments, skip blank lines, perform
                                             arbitrary function before reading
                                             or writing).

    :py:mod:`~plastid.util.io.openers`       Wrappers for opening and closing files
    =======================================  ======================================

"""



#!/usr/bin/env python
"""Tests and validates classes from :py:mod:`plastid.genomics.genome_array`,
these being |GenomeArray|, |SparseGenomeArray| and |BAMGenomeArray|,
using test data found in plastid.test.data. 

This module additionally contains utilites to generate other test datasets.
To do, please see the documentation for :py:func:`create_dataset`. Note,
this requires as external dependencies, Bowtie, Tophat, and Samtools.

"""
import copy
import tempfile
import os
import subprocess
import functools
import re
import unittest
import warnings

import pysam
import numpy.random

from pkg_resources import resource_filename, cleanup_resources
from nose.plugins.attrib import attr

from Bio import SeqIO

import plastid.util.services.exceptions

from plastid.readers.bed import BED_Reader
from plastid.genomics.genome_array import (
    GenomeArray,
    SparseGenomeArray,
    BigWigGenomeArray,
    BAMGenomeArray,
    ThreePrimeMapFactory,
    FivePrimeMapFactory,
    CenterMapFactory,
    five_prime_map,
    three_prime_map,
    center_map,
)
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from plastid.genomics.seqtools import random_seq
from plastid.util.io.openers import NullWriter
from plastid.util.services.decorators import skip_if_abstract
from plastid.util.services.mini2to3 import cStringIO

#===============================================================================
# INDEX: annotations/data used in unit tests and in generation of test datasets
#===============================================================================

# parameters to flesh out unit tests
# these are used by AbstractGenomeArrayHelper.set_class_parameters below
_GENOME_ARRAY_PARAMS = {
    "test_class": GenomeArray,
    "empty_regions": ["splice", "introns"],
    "native_format": "bowtie",
}

_SPARSE_GENOME_ARRAY_PARAMS = {
    "test_class": SparseGenomeArray,
    "empty_regions": ["splice", "introns"],
    "native_format": "bowtie",
}

_BIGWIG_GENOME_ARRAY_PARAMS = {
    "test_class": BigWigGenomeArray,
    "empty_regions": ["introns"],
    "native_format": "bigwig",
}

_BAM_GENOME_ARRAY_PARAMS = {
    "test_class": BAMGenomeArray,
    "empty_regions": ["introns"],
    "native_format": "BAM",
}

# descriptions of mapping configurations that we will use in test datasets
_SAMPLE_PAT = re.compile(r"(center|fiveprime|threeprime)_([0-9]+)")
_SAMPLE_BASES = [
    'center_0',
    'center_12',
    'fiveprime_0',
    'fiveprime_15',
    'threeprime_0',
    'threeprime_15',
]

_GA_MAP_FUNCTIONS = {
    "fiveprime": five_prime_map,
    "threeprime": three_prime_map,
    "center": center_map,
}

_BAM_MAP_RULES = {
    "fiveprime_0": FivePrimeMapFactory(),
    "threeprime_0": ThreePrimeMapFactory(),
    "center_0": CenterMapFactory(),
    "fiveprime_15": FivePrimeMapFactory(15),
    "threeprime_15": ThreePrimeMapFactory(15),
    "center_12": CenterMapFactory(12),
}

# constants/names of files in test datasets, to use in test cases
# or to generate using methods below
#
# all filenames are relative to a base_folder that is passed to individual functions
# these files will be created by create_test_dataset(), below
_TEST_FILES = {
    "variable_step_fw": os.path.join("wig", "variable_step_fw.wig"),
    "variable_step_rc": os.path.join("wig", "variable_step_rc.wig"),
    "bedgraph_fw": os.path.join("wig", "bedgraph_fw.wig"),
    "bedgraph_rc": os.path.join("wig", "bedgraph_rc.wig"),
    "juncs": os.path.join("ebwt", "chrA.juncs"),
    "bowtie_index": os.path.join("ebwt", "chrA"),
    "bed": os.path.join("bed", "chrA.bed"),
    "reads": os.path.join("fasta", "chrA_reads.fa"),
    "bowtie": os.path.join("align", "chrA_unspliced.bowtie"),
    "bam": os.path.join("align", "chrA_tophat.bam"),
    "genome": os.path.join("fasta", "chrA.fa"),
}

# annotation data
TEST_CHR_BED = """chrA    100    1100    unique_plus    0    +    -1    -1    0,0,0    1    1000,    0,
chrA    100    1100    unique_minus    0    -    -1    -1    0,0,0    1    1000,    0,
chrA    1200    2250    entire_repeat_region_plus    0    +    -1    -1    0,0,0    1    1050,    0,
chrA    1200    2250    entire_repeat_region_minus    0    -    -1    -1    0,0,0    1    1050,    0,
chrA    1200    1700    repeat_1_plus    0    +    -1    -1    0,0,0    1    500,    0,
chrA    1200    1700    repeat_1_minus    0    -    -1    -1    0,0,0    1    500,    0,
chrA    1750    2250    repeat_2_plus    0    +    -1    -1    0,0,0    1    500,    0,
chrA    1750    2250    repeat_2_minus    0    -    -1    -1    0,0,0    1    500,    0,
chrA    2350    2475    splice_plus    100    +    -1    -1    0,0,0    2    25,25,    0,100,
chrA    2350    2475    splice_minus    100    -    -1    -1    0,0,0    2    25,25,    0,100,
chrA    2375    2450    intron_plus    0    +    -1    -1    0,0,0    1    75,    0,
chrA    2375    2450    intron_minus    0    -    -1    -1    0,0,0    1    75,    0,""".replace(
    "    ", "\t"
)

TEST_CHR_JUNCS = """chrA    2374    2450    +
chrA    2374    2450    -""".replace("    ", "\t")

# miscellaneous constants
STRAND_KEYS = {"+": "fw", "-": "rc"}
DEFAULT_READS_PER_REGION = 1000
DEFAULT_READ_LENGTH = 30

#===============================================================================
# INDEX: Helper functions for unit tests and test dataset creation methods below
#===============================================================================


def tearDownModule():
    """Remove test dataset files after unit tests are complete"""
    cleanup_resources()


def fetch_regions():
    """Parses test regions of interest for synthetic genomes
    
    Returns
    -------
    list<SegmentChain>
    """
    return list(BED_Reader(cStringIO.StringIO(TEST_CHR_BED), return_type=SegmentChain))


def _read_count_vectors(base_folder):
    """Read count vectors from a synthetic datasets
    generated by :py:method:`create_test_dataset`

    Parameters
    ----------
    base_folder : str
        path to base folder passed to :py:method:`create_test_dataset`
    
    Returns
    -------
    dict : dict of numpy.ndarrays of count data
    """
    dtmp = {}
    for k in _SAMPLE_BASES:
        for strand_key in ("fw", "rc"):
            dtmp["%s_%s" % (k, strand_key)] = numpy.loadtxt(
                os.path.join(base_folder, "count_vectors", "%s_%s.txt" % (k, strand_key))
            )
    return dtmp


def _read_bowtie_files_to_genome_arrays(base_folder, test_class=GenomeArray):
    """Construct |GenomeArray| s from bowtie files

    Parameters
    ----------
    base_folder : str
        path to base folder passed to create_test_dataset()

    test_class : class
        Subclass of |MutableGenomeArray| (e.g. |GenomeArray| or |SparseGenomeArray| to test)

    Returns
    -------
    dict : dict of |GenomeArray| s of mapped read alignments from bowtie
    """
    gnds = {}
    for k in _SAMPLE_BASES:
        mapping, offset = _SAMPLE_PAT.search(k).groups()
        trans_key = "nibble" if mapping == "center" else "offset"
        trans_args = {trans_key: int(offset)}
        gnds[k] = test_class()
        with open(os.path.join(base_folder, _TEST_FILES["bowtie"])) as fh:
            gnds[k].add_from_bowtie(fh, _GA_MAP_FUNCTIONS[k.split("_")[0]], **trans_args)

    return gnds


def _get_ivc_numpy_counts(ivc, count_vec):
    """Fetches appropriately-spliced counts at each position in an ROI from a numpy array
    
    Parameters
    ----------
    ivc : |SegmentChain|
        SegmentChain describing region of interest
    
    count_vec : numpy.ndarray
        numpy.ndarray, in coordinates matching those of ivc
    
    Returns
    -------
    numpy.ndarray : numpy.ndarray of counts each each position in ivc
    """
    counts = []
    for iv in ivc:
        counts.extend(count_vec[iv.start:iv.end])

    if ivc.spanning_segment.strand == "-":
        counts = counts[::-1]

    return numpy.array(counts)


#===============================================================================
# INDEX: unittest suites
#===============================================================================


class AbstractGenomeArrayHelper(unittest.TestCase):
    """Abstract base class for various types of |AbstractGenomeArray| test cases"""

    set_up = False

    @staticmethod
    def set_class_parameters(
            cls, params, test_folder=resource_filename("plastid", "test/data/mini"), tol=1e-8
    ):
        """Set class parameters on the creation of the first instance.
        This is a bit of a hack because we need to set class parameters.
        We can't do this in a ``setUpClass`` method, because ``setUpClass`` only 
        accepts a single parameter (the class). We don't want to do this in
        ``__init__`` either, because unittest calls ``__init__`` once per test run,
        and these operations are expensive. So, instead we define this method,
        and call it from ``__init__`` if and only if ``cls.set_up == False``
        
        Parameters
        ----------
        cls : class
            class that is a subclass of :py:class:`unittest.TestCase`,
            to which parameters will be appended
        
        params : dict
            Parameters specific to the set-up of test suites for specific
            types of GenomeArrays
        
        test_folder : str or :py:class:`Resource`
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests        
        """
        cls.test_folder = test_folder
        cls.tol = tol

        cls.test_class = params["test_class"]
        cls.native_format = params["native_format"]
        cls.count_vecs = _read_count_vectors(cls.test_folder)

        cls.regions = fetch_regions()
        cls.region_classes = {
            "unique": [X for X in cls.regions if "unique" in X.get_name()],
            "repeat":
            [X for X in cls.regions if "entire" not in X.get_name() and "repeat" in X.get_name()],
            "introns": [X for X in cls.regions if "intron" in X.get_name()],
            "splice": [X for X in cls.regions if "splice" in X.get_name()],
            "entire": [X for X in cls.regions if "entire" in X.get_name()],
        }

        cls.region_classes["empty"] = []
        cls.empty_names = []
        for k in params["empty_regions"]:
            my_regions = cls.region_classes[k]
            cls.region_classes["empty"].extend(my_regions)
            cls.empty_names.extend([X.get_name() for X in my_regions])

        cls.expected_unnorm_sum = 0
        #for region in set(cls.regions) - set(cls.region_classes["empty"]) - set(cls.region_classes["entire"]):
        read_regions = [
            X for X in cls.regions
            if all([X.get_name() not in cls.empty_names, "entire" not in X.get_name()])
        ]
        for region in read_regions:
            vec_key = "fw" if region.strand == "+" else "rc"
            cls.expected_unnorm_sum += _get_ivc_numpy_counts(
                region, cls.count_vecs["fiveprime_0_%s" % vec_key]
            ).sum()

        cls.set_up = True

    def __init__(
            self,
            methodName='runTest',
            params={},
            test_folder=resource_filename("plastid", "test/data/mini"),
            tol=1e-8
    ):
        """Initialize test case to run a single method. 
        We override this method to make sure expensive operations are only run when
        the first instance is made, and then stored in class attributes

        Parameters
        ----------
        methodName : str
            Name of method being run. Required by :py:class:`unittest.TestCase`

        params : dict
            Parameters specific to the set-up of test suites for specific
            |AbstractGenomeArray| subclasses. Don't change these
            
        test_folder : str or :py:class:`Resource`
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests
        """
        unittest.TestCase.__init__(self, methodName=methodName)
        # only do setup if __init__ is called by a subclass
        if "Abstract" not in self.__class__.__name__:
            if self.__class__.set_up == False:
                AbstractGenomeArrayHelper.set_class_parameters(
                    self.__class__, params=params, test_folder=test_folder, tol=tol
                )

    @skip_if_abstract
    def test_chroms(self):
        for v in self.gnds.values():
            self.assertEqual(set(v.chroms()), set(["chrA"]))

    @skip_if_abstract
    def test_strands(self):
        possible_strands = set(["+", "-", "."])
        for v in self.gnds.values():
            self.assertGreaterEqual(len(set(v.strands()) & possible_strands), 0)

    @skip_if_abstract
    def test_test_class(self):
        # Assure all genome arrays tested are of correct subclass
        for k, v in self.gnds.items():
            self.assertTrue(
                isinstance(v, self.test_class),
                "Test %s: instance is of wrong class (expected: %s, found %s)" %
                (k, self.test_class.__name__, v.__class__.__name__)
            )

    @skip_if_abstract
    def test_native_import_positionwise_equality_unique_regions(self):
        for k in _SAMPLE_BASES:
            for region in self.region_classes["unique"]:
                strand_key = STRAND_KEYS[region.spanning_segment.strand]
                gnd_counts = numpy.array(region.get_counts(self.gnds[k]))
                self.assertGreater(gnd_counts.sum(), 0, "Region is empty in sample %s" % k)
                known_counts = _get_ivc_numpy_counts(
                    region, self.count_vecs["%s_%s" % (k, strand_key)]
                )

                max_err = max(abs(gnd_counts - known_counts))
                msg1 = "Positionwise count difference (%s) exceeded tolerance (%s) for '%s' file import for sample test '%s'" % (
                    self.tol, max_err, self.native_format, k
                )
                self.assertLessEqual(max_err, self.tol, msg1)

                sum_diff = abs(known_counts.sum() - gnd_counts.sum())
                msg2 = "Error in difference of total counts (%s) exceeded tolerance (%s) for '%s' import for sample test %s" % (
                    sum_diff, self.tol, self.native_format, k
                )
                self.assertLessEqual(sum_diff, self.tol, msg2)

    @skip_if_abstract
    def test_native_import_positionwise_equality_repeat_regions(self):
        # test sums of position-wise vectors for repeat regions
        plus_repeat = [X for X in self.region_classes["repeat"] if X.spanning_segment.strand == "+"]
        minus_repeat = [
            X for X in self.region_classes["repeat"] if X.spanning_segment.strand == "-"
        ]

        lengths = set([X.length for X in plus_repeat + minus_repeat])
        self.assertEqual(len(lengths), 1)

        for k in _SAMPLE_BASES:
            plus_vec = numpy.zeros(plus_repeat[0].length)
            minus_vec = numpy.zeros(plus_repeat[0].length)

            known_plus_vec = numpy.zeros(plus_repeat[0].length)
            known_minus_vec = numpy.zeros(plus_repeat[0].length)

            for region in plus_repeat:
                plus_vec += region.get_counts(self.gnds[k])
                known_plus_vec += _get_ivc_numpy_counts(
                    region, self.count_vecs["%s_%s" % (k, "fw")]
                )

            for region in minus_repeat:
                minus_vec += region.get_counts(self.gnds[k])
                known_minus_vec += _get_ivc_numpy_counts(
                    region, self.count_vecs["%s_%s" % (k, "rc")]
                )

            self.assertGreater(plus_vec.sum(), 0)
            self.assertGreater(minus_vec.sum(), 0)

            self.assertTrue(
                (abs(known_plus_vec - plus_vec) <= self.tol).all(),
                "Positionwise count difference exceeded tolerance %s for %s import on sample test %s on plus strand"
                % (self.tol, self.native_format, k)
            )
            self.assertTrue(
                (abs(known_minus_vec - minus_vec) <= self.tol).all(),
                "Positionwise count difference exceeded tolerance %s for %s import for sample test %s on minus strand"
                % (self.tol, self.native_format, k)
            )

    @skip_if_abstract
    def test_native_import_empty_regions(self):
        # test regions that should be empty (e.g. introns and splicing)
        for k in _SAMPLE_BASES:
            for region in self.region_classes["empty"]:
                self.assertEqual(
                    sum(region.get_counts(self.gnds[k])), 0,
                    "Found counts in region that should be empty for sample test %s" % k
                )

    @skip_if_abstract
    def variablestep_and_bedgraph_export_helper(
            self, wiggle_type, export_function, input_class=None, **kwargs
    ):
        """Helper function to evaluate tests on variable step wiggle or BED export

        Parameters
        ----------
        wiggle_type : str
            Type of wiggle file. "variable_step" or "bedgraph"

        export_function : function
            unbound method defining export type (e.g. GenomeArray.to_variable_step, BAMGenomeArray.to_bedgraph)

        input_class : subclass of |MutableAbstractGenomeArray| or None
            Class into which exported wiggle or bedgraph files will be read. If None, defaults to self.test_class

        kwargs : keyword arguments
        """
        if input_class is None:
            input_class = self.test_class

        for k, v in self.gnds.items():
            fw_out = tempfile.NamedTemporaryFile(mode="w", delete=False)
            rc_out = tempfile.NamedTemporaryFile(mode="w", delete=False)

            export_function(v, fw_out, "test", "+", **kwargs)
            export_function(v, rc_out, "test", "-", **kwargs)

            fw_out.close()
            rc_out.close()

            new_gnd = input_class()
            with open(fw_out.name) as fh:
                new_gnd.add_from_wiggle(fh, "+")
            with open(rc_out.name) as fh:
                new_gnd.add_from_wiggle(fh, "-")

            self.assertGreater(v.lengths()["chrA"], 0)
            self.assertGreater(new_gnd.lengths()["chrA"], 0)

            ivplus = GenomicSegment("chrA", 0, v.lengths()["chrA"], "+")
            ivminus = GenomicSegment("chrA", 0, v.lengths()["chrA"], "-")

            # test equality of what was exported with current state of GenomeArray
            self.assertTrue(
                abs(new_gnd[ivplus] - v[ivplus] <= self.tol).all(),
                "%s wiggle output on plus strand failed positionwise tolerance %s for test %s" %
                (wiggle_type, self.tol, k)
            )
            self.assertGreater(
                new_gnd[ivplus].sum(), 0,
                "No counts found for %s reimport test %s" % (wiggle_type, k)
            )

            self.assertTrue(
                abs(new_gnd[ivminus] - v[ivminus] <= self.tol).all(),
                "%s wiggle output on minus strand failed positionwise tolerance %s for test %s" %
                (wiggle_type, self.tol, k)
            )
            self.assertGreater(
                new_gnd[ivminus].sum(), 0,
                "No counts found for %s reimport test %s" % (wiggle_type, k)
            )

            # ground-truth test against numpy arrays for unique regions
            for region in self.region_classes["unique"]:
                strand_key = STRAND_KEYS[region.spanning_segment.strand]
                gnd_counts = numpy.array(region.get_counts(new_gnd))
                self.assertGreater(
                    gnd_counts.sum(), 0, "Reimported region is empty in sample %s" % k
                )
                known_counts = _get_ivc_numpy_counts(
                    region, self.count_vecs["%s_%s" % (k, strand_key)]
                )
                max_err = max(abs(gnd_counts - known_counts))
                self.assertLessEqual(
                    max_err, self.tol,
                    "Positionwise count difference (%s) exceeded tolerance (%s) for %s reimport after export from class '%s' for sample '%s'"
                    % (self.tol, max_err, self.native_format, self.test_class, k)
                )

            os.remove(fw_out.name)
            os.remove(rc_out.name)

    @skip_if_abstract
    def test_unnormalized_sum(self):
        for k, v in self.gnds.items():
            v.set_normalize(False)
            found = v.sum()
            expected = self.expected_unnorm_sum
            err = abs(found - expected)
            err_msg = "Observed error (%s) in unnormalized sum (observed %s; expected %s) greater than tolerance (%s) for sample '%s'" % (
                err, found, expected, self.tol, k
            )
            self.assertLessEqual(err, self.tol, err_msg)

    @skip_if_abstract
    def test_normalize_not_change_sum(self):
        for k, v in self.gnds.items():
            v.set_normalize(True)
            found_sum = v.sum()
            err_msg = "Normalize flag changed sum to %s from %s for sample '%s'" % (
                found_sum, self.expected_unnorm_sum, k
            )
            err = abs(found_sum - self.expected_unnorm_sum)
            self.assertLessEqual(err, self.tol, err_msg)
            v.set_normalize(False)

    @skip_if_abstract
    def test_set_and_reset_sum(self):
        expected_unnorm_sum2 = 50000
        for k, v in self.gnds.items():
            v.set_sum(expected_unnorm_sum2)
            v.set_normalize(False)

            found = v.sum()
            err = abs(found - expected_unnorm_sum2)
            err_msg = "Observed error (%s) in sample '%s' set unnormalized sum (observed %s; expected %s) greater than tolerance %s" % (
                err, k, found, expected_unnorm_sum2, self.tol
            )
            self.assertLessEqual(err, self.tol, err_msg)

            v.set_normalize(True)
            found = v.sum()
            err = abs(found - expected_unnorm_sum2)
            err_msg = "Observed error (%s) in sample '%s' set normalized sum (observed %s; expected %s) greater than tolerance %s" % (
                err, k, found, expected_unnorm_sum2, self.tol
            )
            self.assertLessEqual(err, self.tol, err_msg)

            v.set_normalize(False)
            v.reset_sum()
            found = v.sum()
            err = abs(found - self.expected_unnorm_sum)
            err_msg = "Observed error (%s) in sample '%s' reset sum (observed %s; expected %s) greater than tolerance %s" % (
                err, k, found, self.expected_unnorm_sum, self.tol
            )
            self.assertLessEqual(err, self.tol, err_msg)

    @skip_if_abstract
    def test_regionwise_normalize_and_sum(self):
        expected_unnorm_sum2 = 50000
        for k, v in self.gnds.items():
            v.set_normalize(False)
            v.reset_sum()

            # add an order of magnitude to account for summing
            tol = self.tol * 10

            # exclude repeat regions, because those will align differently than they were generated
            # remove "empty" also, because this will include spliced regions for some tests,
            # as necessary
            nonrepeat_nonempty = [
                X for X in self.regions
                if all(["repeat" not in X.get_name(),
                        X.get_name() not in self.empty_names])
            ]
            for region in nonrepeat_nonempty:  #set(self.regions) - set(self.region_classes["repeat"]) - set(self.region_classes["empty"]):
                # Make sure baseline number is ok
                found_region_sum = sum(region.get_counts(v))
                expected_region_unnorm = _get_ivc_numpy_counts(
                    region, self.count_vecs["%s_%s" % (k, STRAND_KEYS[region.strand])]
                ).sum()
                err = abs(found_region_sum - expected_region_unnorm)
                self.assertLessEqual(
                    err, tol,
                    "Found unnormalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_unnorm, tol)
                )

                # Test normalize
                v.set_normalize(True)
                expected_region_norm = float(expected_region_unnorm) / v.sum() * 10**6
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_norm)
                self.assertLessEqual(
                    err, tol,
                    "Found normalized region sum (%s) different from expected (%s) more than error (observed %s; tolerance %s) for sample '%s'"
                    % (found_region_sum, expected_region_norm, err, tol, k)
                )

                # Test reversibility
                v.set_normalize(False)
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_unnorm)
                self.assertLessEqual(
                    err, tol,
                    "Found re-unnormalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_unnorm, tol)
                )

                # Set sum, no normalization
                v.set_sum(expected_unnorm_sum2)
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_unnorm)
                self.assertLessEqual(
                    err, tol,
                    "Found post-global-sum-set unnormalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_unnorm, tol)
                )

                # Add normalization on top of set sum
                v.set_normalize(True)
                expected_region_norm2 = float(expected_region_unnorm) / expected_unnorm_sum2 * 10**6
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_norm2)
                self.assertLessEqual(
                    err, tol,
                    "Found post-global-sum-set normalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_norm2, tol)
                )

                # Reset sum, keep normalization
                v.reset_sum()
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_norm)
                self.assertLessEqual(
                    err, tol,
                    "Found post-reset normalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_norm, tol)
                )

                # Revert all
                v.set_normalize(False)
                found_region_sum = sum(region.get_counts(v))
                err = abs(found_region_sum - expected_region_unnorm)
                self.assertLessEqual(
                    err, tol,
                    "Found unnormalized region sum %s different from expected %s more than error %s"
                    % (found_region_sum, expected_region_unnorm, tol)
                )

    @skip_if_abstract
    def test_get_genomicsegment_roi_order_false(self):
        k = _SAMPLE_BASES[0]
        for region in self.region_classes["unique"]:
            seg = region.spanning_segment
            strand_key = STRAND_KEYS[region.spanning_segment.strand]
            #gnd_counts   = self.gnds[k].__getitem__(seg,roi_order=False)
            gnd_counts = self.gnds[k].get(seg, roi_order=False)
            known_counts = self.count_vecs["%s_%s" % (k, strand_key)][seg.start:seg.end]

            max_err = max(abs(gnd_counts - known_counts))
            self.assertLessEqual(
                max_err, self.tol,
                "Positionwise count difference '%s' exceeded tolerance '%s' for %s __getitem__ with roi_order==False for sample test %s"
                % (self.tol, max_err, self.native_format, k)
            )

    @skip_if_abstract
    def test_get_genomicsegment_roi_order_true(self):
        k = _SAMPLE_BASES[0]
        for region in self.region_classes["unique"]:
            seg = region.spanning_segment
            strand_key = STRAND_KEYS[region.spanning_segment.strand]
            gnd_counts = self.gnds[k].get(seg, roi_order=True)
            known_counts = self.count_vecs["%s_%s" % (k, strand_key)][seg.start:seg.end]
            if seg.strand == "-":
                known_counts = known_counts[::-1]

            max_err = max(abs(gnd_counts - known_counts))
            self.assertLessEqual(
                max_err, self.tol,
                "Positionwise count difference '%s' exceeded tolerance '%s' for %s __getitem__ with roi_order==True for sample test %s"
                % (self.tol, max_err, self.native_format, k)
            )

    @skip_if_abstract
    def test_getitem_genomicsegment(self):
        k = _SAMPLE_BASES[0]
        for region in self.region_classes["unique"]:
            seg = region.spanning_segment
            strand_key = STRAND_KEYS[region.spanning_segment.strand]
            gnd_counts = self.gnds[k].__getitem__(seg)
            known_counts = self.count_vecs["%s_%s" % (k, strand_key)][seg.start:seg.end]
            if seg.strand == "-":
                known_counts = known_counts[::-1]

            max_err = max(abs(gnd_counts - known_counts))
            self.assertLessEqual(
                max_err, self.tol,
                "Positionwise count difference '%s' exceeded tolerance '%s' for %s __getitem__ with roi_order==True for sample test %s"
                % (self.tol, max_err, self.native_format, k)
            )

    @skip_if_abstract
    def test_getitem_segmentchain(self):
        k = _SAMPLE_BASES[0]
        for region in self.region_classes["unique"]:
            strand_key = STRAND_KEYS[region.spanning_segment.strand]
            gnd_counts = self.gnds[k][region]  # test
            self.assertGreater(gnd_counts.sum(), 0, "Region is empty in sample %s" % k)
            known_counts = _get_ivc_numpy_counts(region, self.count_vecs["%s_%s" % (k, strand_key)])
            max_err = max(abs(gnd_counts - known_counts))
            self.assertLessEqual(
                max_err, self.tol,
                "Positionwise count difference '%s' exceeded tolerance '%s' for %s import for sample test %s"
                % (self.tol, max_err, self.native_format, k)
            )


class AbstractExportableGenomeArrayHelper(AbstractGenomeArrayHelper):
    @skip_if_abstract
    def test_variablestep_export(self):
        self.variablestep_and_bedgraph_export_helper(
            "variable_step", self.test_class.to_variable_step
        )

    @skip_if_abstract
    def test_bedgraph_export(self):
        self.variablestep_and_bedgraph_export_helper("bedgraph", self.test_class.to_bedgraph)


@attr(test="unit")
@attr(speed="slow")
class TestGenomeArray(AbstractExportableGenomeArrayHelper):
    """Test case for :py:class:`GenomeArray`"""

    set_up = False
    has_gnds = False

    def __init__(
            self,
            methodName='runTest',
            params=_GENOME_ARRAY_PARAMS,
            test_folder=resource_filename("plastid", "test/data/mini"),
            tol=1e-8
    ):
        """Initialize test case to run a single method. 
        We override this method to make sure expensive operations are only run when
        the first instance is made, and then stored in class attributes

        Parameters
        ----------
        methodName : str
            Name of method being run. Required by :py:class:`unittest.TestCase`
        
        params : dict
            Parameters specific to the set-up of test suites for specific
            |AbstractGenomeArray| subclasses. Don't change these
            
        test_folder : str or Resource
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests
        """
        AbstractGenomeArrayHelper.__init__(
            self, methodName=methodName, params=params, test_folder=test_folder, tol=tol
        )
        if self.__class__.has_gnds == False:
            self.__class__.gnds = _read_bowtie_files_to_genome_arrays(
                self.test_folder, self.test_class
            )
            self.__class__.has_gnds = True
            #TestGenomeArray.setUpClassOnlyOnce()

    def test_setitem_genomicsegment_scalar(self):
        ga = GenomeArray({"chrA": 2000})
        segplus = GenomicSegment("chrA", 50, 100, "+")
        segminus = GenomicSegment("chrA", 50, 100, "-")

        # scalar set
        ga[segplus] = 52
        ga[segminus] = 342

        self.assertTrue(
            (ga._chroms["chrA"]["+"][50:100] == 52).all(),
            "%s failed scalar genomicsegment __setitem__ for plus strand."
        )
        self.assertTrue(
            (ga._chroms["chrA"]["-"][50:100] == 342).all(),
            "%s failed scalar genomicsegment __setitem__ for minus strand."
        )
        self.assertEqual(ga.sum(), 52 * len(segplus) + 342 * len(segminus))

    def test_setitem_genomicsegment_vector(self):
        ga = GenomeArray({"chrA": 2000})
        segplus = GenomicSegment("chrA", 50, 100, "+")
        segminus = GenomicSegment("chrA", 50, 100, "-")

        # vector set
        r1 = numpy.random.randint(0, high=242, size=50)
        r2 = numpy.random.randint(0, high=242, size=50)
        ga[segplus] = r1
        ga[segminus] = r2

        self.assertTrue(
            (ga._chroms["chrA"]["+"][50:100] == r1).all(),
            "%s failed vector genomicsegment __setitem__ for plus strand."
        )
        self.assertTrue(
            (ga._chroms["chrA"]["-"][50:100] == r2[::-1]).all(),
            "%s failed vector genomicsegment __setitem__ for minus strand."
        )

        self.assertEqual(ga.sum(), r1.sum() + r2.sum())

    def test_setitem_segmentchain_scalar(self):
        ga = GenomeArray({"chrA": 2000})
        pluschain = SegmentChain(
            GenomicSegment("chrA", 50, 100, "+"),
            GenomicSegment("chrA", 150, 732, "+"),
            GenomicSegment("chrA", 1800, 2500, "+"),
        )
        minuschain = SegmentChain(
            GenomicSegment("chrA", 50, 100, "-"),
            GenomicSegment("chrA", 150, 732, "-"),
            GenomicSegment("chrA", 1800, 2500, "-"),
        )
        ga[pluschain] = 31
        ga[minuschain] = 424
        for seg in pluschain:
            self.assertTrue((ga._chroms[seg.chrom][seg.strand][seg.start:seg.end] == 31).all())

        for seg in minuschain:
            self.assertTrue((ga._chroms[seg.chrom][seg.strand][seg.start:seg.end] == 424).all())

        self.assertEqual(ga.sum(), 31 * pluschain.length + 424 * minuschain.length)

    def test_setitem_segmentchain_vector(self):
        ga = GenomeArray({"chrA": 2000})
        pluschain = SegmentChain(
            GenomicSegment("chrA", 50, 100, "+"),
            GenomicSegment("chrA", 150, 732, "+"),
            GenomicSegment("chrA", 1800, 2500, "+"),
        )
        minuschain = SegmentChain(
            GenomicSegment("chrA", 50, 100, "-"),
            GenomicSegment("chrA", 150, 732, "-"),
            GenomicSegment("chrA", 1800, 2500, "-"),
        )

        plusvec = numpy.random.randint(0, high=250, size=pluschain.length)
        minusvec = numpy.random.randint(0, high=250, size=minuschain.length)

        ga[pluschain] = plusvec
        ga[minuschain] = minusvec

        x = 0
        for seg in pluschain:
            subvec = ga._chroms["chrA"]["+"][seg.start:seg.end]
            self.assertTrue((subvec == plusvec[x:x + len(subvec)]).all())
            x += len(subvec)

        x = 0
        for seg in minuschain:
            subvec = ga._chroms["chrA"]["-"][seg.start:seg.end][::-1]
            self.assertTrue(
                (subvec == minusvec[len(minusvec) - x - len(subvec):len(minusvec) - x]).all()
            )
            x += len(subvec)

        self.assertEqual(ga.sum(), plusvec.sum() + minusvec.sum())

    def variablestep_and_bed_import_helper(self, wiggle_type):
        """Helper function to evaluate tests on variable step wiggle or BEDgraph import

        Parameters
        ----------
        wiggle_type : str
            Type of wiggle file. "variable_step" or "bedgraph"
        """
        gnd = self.test_class()
        with open(os.path.join(self.test_folder, _TEST_FILES["%s_%s" % (wiggle_type, "fw")])) as fh:
            gnd.add_from_wiggle(fh, "+")
        with open(os.path.join(self.test_folder, _TEST_FILES["%s_%s" % (wiggle_type, "rc")])) as fh:
            gnd.add_from_wiggle(fh, "-")

        # Make sure imported counts are nonzero
        self.assertGreater(gnd.sum(), 0, "Import of %s yielded no counts!" % wiggle_type)
        chrA_len = gnd.lengths()["chrA"]

        for strand, trackstub, label in [("+", "fw", "plus"), ("-", "rc", "minus")]:
            my_vec = self.count_vecs["fiveprime_0_%s" % trackstub]
            vec_len = len(my_vec)

            empty_iv = GenomicSegment("chrA", vec_len, chrA_len, strand)
            nonempty_iv = GenomicSegment("chrA", 0, vec_len, strand)
            nonempty_vec = gnd.get(nonempty_iv, roi_order=False)

            # make sure count vector has requisite counts
            self.assertGreater(my_vec.sum(), 0)

            # make sure sums are equal
            self.assertEquals(my_vec.sum(), nonempty_vec.sum())

            # make sure all regions after count vector are empty
            self.assertEquals(
                gnd[empty_iv].sum(), 0, "Found counts in region that should be empty."
            )

            # make sure all positions in regions up to count vector are correct
            max_err = abs(my_vec - nonempty_vec).max()
            self.assertLessEqual(
                max_err, self.tol,
                "Positionwise count difference %s exceeded tolerance %s for wiggle import on %s strand"
                % (max_err, self.tol, label)
            )

    def test_bedgraph_import(self):
        self.variablestep_and_bed_import_helper("bedgraph")

    def test_variablestep_import(self):
        self.variablestep_and_bed_import_helper("variable_step")

    def test_genome_wide_scalar_plus_equals_sum(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        self.assertEqual(len(gnd), expected_length)

        gnd += 5
        self.assertEqual(gnd.sum(), 5 * expected_length)

        gnd -= 5
        self.assertEqual(gnd.sum(), 0)

    def test_genome_wide_scalar_plus_equals_not_change_length(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        self.assertEqual(len(gnd), expected_length)

        gnd += 5
        self.assertEqual(len(gnd), expected_length)

    def test_genome_wide_scalar_plus_sum(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        self.assertEqual(len(gnd), expected_length)

        # add scalar
        gnd2 = gnd + 5
        self.assertEqual(gnd2.sum(), 5 * expected_length)
        self.assertEqual(gnd.sum(), 0)

        # add scalar to occupied gnd
        gnd3 = gnd2 + 1
        self.assertEqual(gnd3.sum(), 6 * expected_length)

    def test_genome_scalar_times_equals_sum(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        self.assertEqual(len(gnd), expected_length)

        gnd *= 5
        self.assertEqual(gnd.sum(), 0)

        gnd += 1
        gnd *= 5
        self.assertEqual(gnd.sum(), 5 * expected_length)

    def test_genome_wide_scalar_times_equals_not_change_length(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        self.assertEqual(len(gnd), expected_length)

        gnd *= 5
        self.assertEqual(len(gnd), expected_length)

    def test_genome_wide_scalar_times(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        expected_length = 2 * sum(chroms.values())

        gnd = self.test_class(chroms)
        gnd += 1
        self.assertEqual(gnd.sum(), expected_length)

        gnd2 = gnd * 2
        self.assertEqual(gnd2.sum(), 2 * expected_length)
        self.assertEqual(gnd.sum(), expected_length)

    def test_genome_wide_array_add_same_size(self):
        chroms = {"chrA": 1000, "chrB": 10000}

        gnd1 = self.test_class(chroms)
        gnd2 = self.test_class(chroms)

        gnd1 += 1
        gnd2 += 3
        gnd3 = gnd1 + gnd2
        self.assertEquals(gnd3.sum(), gnd2.sum() + gnd1.sum())

        iv1plus = GenomicSegment("chrA", 0, 1000, "+")
        iv1minus = GenomicSegment("chrA", 0, 1000, "+")

        self.assertTrue((gnd3[iv1plus] == gnd2[iv1plus] + gnd1[iv1plus]).all())
        self.assertTrue((gnd3[iv1minus] == gnd2[iv1minus] + gnd1[iv1minus]).all())

    def test_genome_wide_array_multiply_same_size(self):
        chroms = {"chrA": 1000, "chrB": 10000}

        gnd1 = self.test_class(chroms)
        gnd2 = self.test_class(chroms)

        gnd1 += 2
        gnd2 += 3
        gnd3 = gnd1 * gnd2

        iv1plus = GenomicSegment("chrA", 0, 1000, "+")
        iv1minus = GenomicSegment("chrA", 0, 1000, "+")

        self.assertTrue((gnd3[iv1plus] == gnd2[iv1plus] * gnd1[iv1plus]).all())
        self.assertTrue((gnd3[iv1minus] == gnd2[iv1minus] * gnd1[iv1minus]).all())

    def test_iadd_no_normalize(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        gnd = self.test_class(chroms)
        gnd.set_normalize(False)
        self.assertEqual(0, gnd.sum())

        iv1plus = GenomicSegment("chrA", 0, 1000, "+")
        iv2plus = GenomicSegment("chrA", 0, 500, "+")
        gnd[iv1plus] += 1
        self.assertEqual(1000, gnd.sum())

        gnd[iv1plus] += 1
        self.assertEqual(2000, gnd.sum())

        gnd[iv2plus] += 1
        self.assertEqual(2500, gnd.sum())

    def test_iadd_with_normalize_raises_warning(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        gnd = self.test_class(chroms)
        gnd.set_normalize(True)

        def my_func(ga):
            ga.set_normalize(True)
            ga[GenomicSegment("chrA", 0, 1000, "+")] += 5

        # manually reset registry before test
        plastid.util.services.exceptions.pl_once_registry = {}
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            my_func(gnd)

        got_warning = False
        for w in warns:
            if "turning off normalization" in str(w.message):
                got_warning = True

        self.assertTrue(got_warning)

    def test_eq(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        gnd1 = self.test_class(chroms)
        gnd2 = self.test_class(chroms)

        self.assertEqual(gnd1, gnd2)

        # diff chrom nonzero positions
        gnd2[GenomicSegment("chrC", 500, 1000, "-")] = 200
        self.assertNotEqual(gnd1, gnd2)

        # same chroms, nonzero positions, and values
        gnd1[GenomicSegment("chrC", 500, 1000, "-")] = 200
        self.assertEqual(gnd1, gnd2)

        # same chroms and nonzero positions, diff values
        gnd1[GenomicSegment("chrC", 500, 1000, "-")] += 200
        self.assertNotEqual(gnd1, gnd2)

    def test_setters_and_getters(self):
        chroms = {"chrA": 1000, "chrB": 10000}
        gnd = self.test_class(chroms)

        #generate random read data set
        for chr_name, chr_len in chroms.items():
            for strand in ("+", "-"):
                num_iters = numpy.random.randint(50, 100)  #000)
                starts = numpy.random.randint(0, chr_len - 50, size=num_iters)
                ends = [X + numpy.random.randint(0, 50) for X in starts]
                vals = numpy.random.randint(0, 100, size=num_iters)
                for i in range(num_iters):
                    iv = GenomicSegment(chr_name, int(starts[i]), int(ends[i]), strand)
                    gnd[iv] = vals[i]

        self.assertEquals(
            len(gnd), 2 * sum(chroms.values()),
            "Chromosome lengths incorrect: %s vs %s" % (len(gnd), 2 * sum(chroms.values()))
        )

        # auto-grow
        iv1 = GenomicSegment("chrA", 10000, 11000, "+")
        iv2 = GenomicSegment("chrA", 10500, 11000, "+")
        iv3 = GenomicSegment("chrA", int(5 * 1e5) + 10500, int(5 * 1e5) + 11000, "+")
        iv4 = GenomicSegment("chrB", int(5 * 1e5) + 10500, int(5 * 1e5) + 11000, "+")
        gnd[iv1] = 1
        self.assertEquals(sum(gnd[iv1]), 1000, "Auto-grow failed during set")
        gnd[iv2] += 1
        self.assertEquals(sum(gnd[iv1]), 1500, "+= failed during set")
        self.assertEquals(sum(gnd[iv2]), 1000)

        self.assertGreater(gnd.lengths()["chrA"], 1000, "Auto-grow failed chrA")

        gnd[iv3] += 1
        self.assertEqual(sum(gnd[iv3]), 500, "+= failed during set")
        self.assertEqual(sum(gnd[iv4]), 0, "Counts where not expected")

        # setters & getters
        iv1 = GenomicSegment("chrA", 200, 500, "+")
        oldvals = copy.deepcopy(gnd[iv1])
        gnd[iv1] = 5
        newvals = copy.deepcopy(gnd[iv1])
        self.assertTrue((oldvals != newvals).any(), "Set failed")
        self.assertEqual(newvals.sum(), 5 * len(newvals))
        self.assertEqual(newvals.sum(), 5 * len(iv1))

        newvals2 = copy.deepcopy(gnd[iv1])
        self.assertTrue((newvals2 == newvals).all(), "Set failed")

        newvals = newvals2 = None
        gnd[iv1] = oldvals
        new_old = gnd[iv1]
        self.assertTrue((new_old == oldvals).all(), "Set valed")

        # scalar add
        gnd_plus5 = gnd + 5
        for iv in (iv1, iv2, iv3, iv4):
            self.assertTrue(
                (gnd_plus5[iv] == (gnd[iv] + 5)).all(),
                "Scalar addition globally failed interval test"
            )

        # FIXME- what is len of a sparse array?
        diff = abs(gnd_plus5.sum() - (gnd.sum() + 5 * len(gnd)))
        self.assertLess(
            diff, self.tol,
            "Error in genome-wide scalar addition (%s) exceeded tolerance %s. Raw vals: %s vs %s " %
            (diff, self.tol, gnd_plus5.sum(), gnd.sum() + 5 * len(gnd))
        )

        # scalar multiply
        gnd3 = gnd * 3
        for iv in (iv1, iv2, iv3, iv4):
            self.assertTrue(
                (gnd3[iv] == (gnd[iv] * 3)).all(),
                "Scalar multip.lication globally failed interal test"
            )

        quotient = 1.0 - (gnd3.sum() / 3 * gnd.sum())
        self.assertLess(
            quotient, self.tol,
            "Error in scalar multiplication (%s) exceeded tolerance %s" % (diff, self.tol)
        )

        # genome-wide multiply
        gnd4 = gnd + gnd3
        gndmul = gnd3 * gnd4
        for iv in (iv1, iv2, iv3, iv4):
            self.assertTrue(
                (gndmul[iv] - (gnd3[iv] * gnd4[iv]) <= self.tol).all(),
                "Error in genome-wide multiply failed exceeded tolerance %s" % self.tol
            )

        # genome-wide add
        is_ok = True
        for iv in (iv1, iv2, iv3):
            is_ok &= sum(gnd4[iv]) > 0
            self.assertTrue(
                (gnd4[iv] == gnd[iv] + gnd3[iv]).all(),
                "Error in genome-wide addition exceeded tolerance %s" % self.tol
            )
            self.assertGreater(sum(gnd4[iv]), 0)

    def test_nonzero(self):
        excluded_set = set([])
        for region in self.region_classes["repeat"] + self.region_classes["splice"]:
            excluded_set |= region.get_position_set()

        for k, v in self.gnds.items():
            nz_dict = v.nonzero()
            for strand in v.strands():
                strand_key = STRAND_KEYS[strand]
                expected = self.count_vecs["%s_%s" % (k, strand_key)].nonzero()[0]

                found = nz_dict["chrA"][strand]
                self.assertEqual(set(expected) - excluded_set, set(found) - excluded_set)


@attr(test="unit")
@attr(speed="slow")
class TestSparseGenomeArray(TestGenomeArray):
    """Test suite for |SparseGenomeArray|"""

    set_up = False
    has_gnds = False

    def __init__(
            self,
            methodName='runTest',
            params=_SPARSE_GENOME_ARRAY_PARAMS,
            test_folder=resource_filename("plastid", "test/data/mini"),
            tol=1e-8
    ):
        """Initialize test case to run a single method. 
        We override this method to make sure expensive operations are only run when
        the first instance is made, and then stored in class attributes

        Parameters
        ----------
        methodName : str
            Name of method being run. Required by :py:class:`unittest.TestCase`
        
        params : dict
            Parameters specific to the set-up of test suites for specific
            AbstractgenomeArray subclasses. Don't change these
            
        test_folder : str or Resource
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests
        """
        TestGenomeArray.__init__(
            self, methodName=methodName, params=params, test_folder=test_folder, tol=tol
        )


@attr(test="unit")
@attr(speed="slow")
class TestBigWigGenomeArray(AbstractGenomeArrayHelper):
    """Test suite for |SparseGenomeArray|"""

    set_up = False
    has_gnds = False

    def __init__(
            self,
            methodName='runTest',
            params=_BIGWIG_GENOME_ARRAY_PARAMS,
            test_folder=resource_filename("plastid", "test/data/mini"),
            tol=1e-3
    ):
        """Initialize test case to run a single method. 
        We override this method to make sure expensive operations are only run when
        the first instance is made, and then stored in class attributes

        Parameters
        ----------
        methodName : str
            Name of method being run. Required by :py:class:`unittest.TestCase`
        
        params : dict
            Parameters specific to the set-up of test suites for specific
            AbstractgenomeArray subclasses. Don't change these
            
        test_folder : str or Resource
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests
        """
        AbstractGenomeArrayHelper.__init__(
            self, methodName=methodName, params=params, test_folder=test_folder, tol=tol
        )
        if self.__class__.has_gnds == False:
            TestBigWigGenomeArray.gnds = TestBigWigGenomeArray.read_bigwig_files()
            self.__class__.has_gnds = True

    @staticmethod
    def read_bigwig_files():
        """Read bigwig files into a dictionary
        """
        dtmp = {}
        for k in _SAMPLE_BASES:
            dtmp[k] = ga = BigWigGenomeArray(fill=0)

            fw = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_%s_fw.bw" % k)
            rc = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_%s_rc.bw" % k)

            ga.add_from_bigwig(fw, "+")
            ga.add_from_bigwig(rc, "-")

        return dtmp

    def test_multiple_same_strand_sum(self):
        # should see sum double
        bigwigfile = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_center_12_fw.bw")
        bw = BigWigGenomeArray(fill=0)

        bw.add_from_bigwig(bigwigfile, "+")
        self.assertLessEqual(abs(bw.sum() - 4000), self.tol)

        bw.add_from_bigwig(bigwigfile, "+")
        self.assertLessEqual(abs(bw.sum() - 8000), self.tol)

        bw.add_from_bigwig(bigwigfile, "+")
        self.assertLessEqual(abs(bw.sum() - 12000), self.tol)

    def test_multiple_same_strand_fetch(self):
        bigwigfw = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_center_12_fw.bw")
        bigwigrc = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_center_12_rc.bw")
        wigfw = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_center_12_fw.wig")
        wigrc = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_center_12_rc.wig")

        bw = BigWigGenomeArray(fill=0)
        bw.add_from_bigwig(bigwigfw, "+")
        bw.add_from_bigwig(bigwigfw, "+")
        bw.add_from_bigwig(bigwigrc, "-")
        bw.add_from_bigwig(bigwigrc, "-")

        ga = GenomeArray(bw.lengths())
        with open(wigfw) as fh:
            ga.add_from_wiggle(fh, "+")
        with open(wigrc) as fh:
            ga.add_from_wiggle(fh, "-")

        for chrom, length in bw.lengths().items():
            for strand in bw.strands():
                seg = GenomicSegment(chrom, 0, length, strand)
                maxdiff = abs(bw[seg] - 2 * ga[seg]).max()
                msg = "Maximum difference for multiple_strand_fetch (%s) exceeds tolerance (%s)" % (
                    maxdiff, self.tol
                )
                self.assertLessEqual(maxdiff, self.tol, msg)

    def test_to_genome_array(self):
        for test, orig in self.gnds.items():
            fw = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_%s_fw.wig" % test)
            rc = os.path.join(TestBigWigGenomeArray.test_folder, "wig", "bw_%s_rc.wig" % test)

            expected = GenomeArray()
            with open(fw) as fh:
                expected.add_from_wiggle(fh, "+")
            with open(rc) as fh:
                expected.add_from_wiggle(fh, "-")

            found = orig.to_genome_array()

            for chrom, length in expected.lengths().items():
                for strand in ("+", "-"):
                    seg = GenomicSegment(chrom, 0, length, strand)
                    diffvec = abs(orig[seg] - found[seg])
                    diffmax = diffvec.max()
                    msg1 = "Maximum difference between exported GenomeArray and BigWigGenomeArray (%s) exceeds tolerance (%s) for test '%s' strand '%s'" % (
                        diffmax, self.tol, test, strand
                    )
                    self.assertLessEqual(diffmax, self.tol, msg1)

            for chrom, length in expected.lengths().items():
                for strand in ("+", "-"):
                    seg = GenomicSegment(chrom, 0, length, strand)
                    diffvec = abs(expected[seg] - found[seg])
                    diffmax = diffvec.max()
                    msg1 = "Maximum difference between exported GenomeArray and wiggle-imported array (%s) exceeds tolerance (%s) for test '%s' strand '%s'" % (
                        diffmax, self.tol, test, strand
                    )

                    self.assertLessEqual(diffmax, self.tol, msg1)


class FakeDict(object):
    """Creates a dictionary-like object that provies dictionary-like access
    to a BAMGenomeArray under various mapping rules, as if it were a collection
    of separate GenomeArrays. This is only a convenience class to allow us to
    re-use functions in the |AbstractGenomeArrayHelper| test suite in
    |TestBAMGenomeArray|
    """

    def __init__(self, bga, map_functions=_BAM_MAP_RULES):
        """Create a FakeDict

        Parameters
        ----------
        bga : |BAMGenomeArray|

        map_functions : dict
            Dictionary mapping descriptive names to mapping functions,
            such as those made by :py:func:`plastid.genomics.genome_array.FivePrimeMapFactory`
        """
        self.bga = bga
        self.map_functions = map_functions

    def __getitem__(self, key):
        self.bga.set_mapping(self.map_functions[key])
        return self.bga

    def items(self):
        for k in self.map_functions:
            yield (k, self[k])

    def values(self):
        # must use key, to trigger map setting in __getitem__
        for k in self.map_functions:
            yield self[k]


@attr(test="unit")
@attr(speed="slow")
class TestBAMGenomeArray(AbstractExportableGenomeArrayHelper):

    set_up = False
    has_gnds = False

    def __init__(
            self,
            methodName='runTest',
            params=_BAM_GENOME_ARRAY_PARAMS,
            test_folder=resource_filename("plastid", "test/data/mini"),
            tol=1e-8
    ):
        """Initialize test case to run a single method. 
        We override this method to make sure expensive operations are only run when
        the first instance is made, and then stored in class attributes

        Parameters
        ----------
        methodName : str
            Name of method being run. Required by py:class:`unittest.TestCase`

        params : dict
            Parameters specific to the set-up of test suites for specific
            AbstractgenomeArray subclasses. Don't change these
            
        test_folder : str or Resource
            Real or virtual location of folder of test data

        tol : float
            Tolerance for numerical differences between expected and observed
            values in the various tests
        """
        AbstractGenomeArrayHelper.__init__(
            self, methodName=methodName, params=params, test_folder=test_folder, tol=tol
        )
        if self.__class__.has_gnds == False:
            bga = BAMGenomeArray(
                [pysam.Samfile(os.path.join(self.test_folder, _TEST_FILES["bam"]), "rb")]
            )
            TestBAMGenomeArray.gnds = FakeDict(bga)
            TestBAMGenomeArray.bga = bga
            self.__class__.has_gnds = True

    def test_open_str_filename(self):
        z = BAMGenomeArray(os.path.join(self.test_folder, _TEST_FILES["bam"]))
        self.assertEqual(z.sum(), self.bga.sum())

    def test_open_multi_list(self):
        v = [os.path.join(self.test_folder, _TEST_FILES["bam"])] * 2
        z = BAMGenomeArray(v)
        self.assertEqual(z.sum(), 2 * self.bga.sum())

    def test_open_multi_filename(self):
        f = os.path.join(self.test_folder, _TEST_FILES["bam"])
        z = BAMGenomeArray(f, f)
        self.assertEqual(z.sum(), 2 * self.bga.sum())

    def mutable_conversion_helper(self, new_class):
        """Helper function to test conversion of |BAMGenomeArray| to various |MutableAbstractGenomeArray| types

        Parameters
        ----------
        new_class : class
            Non-abstract subclass of |MutableAbstractGenomeArray|
        """
        ivplus = GenomicSegment("chrA", 0, self.bga.lengths()["chrA"], "+")
        ivminus = GenomicSegment("chrA", 0, self.bga.lengths()["chrA"], "-")
        for k, v in self.gnds.items():
            new_gnd = v.to_genome_array(array_type=new_class)

            for iv in (ivplus, ivminus):

                self.assertGreater(v[iv].sum(), 0)
                self.assertGreater(new_gnd[iv].sum(), 0)

                max_err = max(abs(v[iv] - new_gnd[iv]))
                err_message = "%s BAMGenomeArray conversion to %s error %s exceeded tolerance %s." % (
                    k, new_class.__name__, max_err, self.tol
                )
                self.assertLess(max_err, self.tol, err_message)

    def test_to_genome_array(self):
        self.mutable_conversion_helper(GenomeArray)

    def test_to_sparse_genome_array(self):
        self.mutable_conversion_helper(SparseGenomeArray)

    def variablestep_and_bedgraph_export_helper(self, wiggle_type, export_function):
        # override function so we can test window size parameters in export
        for window_size in (1, 2, 5, 10, 25, 100, 500, 1000, 10000):
            AbstractGenomeArrayHelper.variablestep_and_bedgraph_export_helper(
                self,
                wiggle_type,
                export_function,
                input_class=GenomeArray,
                window_size=window_size
            )

    def test_add_remove_filter(self):
        # add a plus-strand filter and require minus strand regions be zero
        # then remove and watch it come back
        bga = self.bga

        def minus_only_filter(read):
            return read.is_reverse

        entire_iv_plus = GenomicSegment("chrA", 0, bga.lengths()["chrA"], "+")
        entire_iv_minus = GenomicSegment("chrA", 0, bga.lengths()["chrA"], "-")

        # fetch counts & check
        pre_plus = bga[entire_iv_plus]
        self.assertGreater(pre_plus.sum(), 0)

        pre_minus = bga[entire_iv_minus]
        self.assertGreater(pre_minus.sum(), 0)

        # add filter, re-fetch
        bga.add_filter("minus_only", minus_only_filter)
        post_plus = bga[entire_iv_plus]
        post_minus = bga[entire_iv_minus]

        self.assertEqual(post_plus.sum(), 0)
        self.assertFalse((post_plus == pre_plus).all())

        self.assertEqual(post_minus.sum(), pre_minus.sum())
        self.assertTrue((post_minus == pre_minus).all())

        # remove_filter, re_fetch
        bga.remove_filter("minus_only")
        post_post_plus = bga[entire_iv_plus]
        post_post_minus = bga[entire_iv_minus]

        self.assertEqual(post_post_plus.sum(), pre_plus.sum())
        self.assertTrue((post_post_plus == pre_plus).all())

        self.assertEqual(post_post_minus.sum(), pre_minus.sum())
        self.assertTrue((post_post_minus == pre_minus).all())


#===============================================================================
# INDEX: tools for generating test datasets with known results
#===============================================================================


def _detect_or_create_folders(base_folder):
    """Creates and tests folder hierarchy needed for unit/integrative tests below.

    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created
    """
    sub_folders = ["fasta", "ebwt", "count_vectors", "align", "wig", "bed"]
    if not os.path.isdir(base_folder):
        os.mkdir(base_folder)
    for name in sub_folders:
        sf = os.path.join(base_folder, name)
        if not os.path.isdir(sf):
            os.mkdir(sf)

    # BED file for use later
    with open(os.path.join(base_folder, _TEST_FILES["bed"]), "w") as fout:
        fout.write(TEST_CHR_BED)
        fout.close()

    # .juncs file for tophat
    with open(os.path.join(base_folder, _TEST_FILES["juncs"]), "w") as fout:
        fout.write(TEST_CHR_JUNCS)
        fout.close()


def detect_base_folder(func):
    """Decorator function to ensure that folders required by functions below exist.
    For this decorator to work, the function it wraps MUST require base_folder
    as its first parameter.

    Parameters
    ----------
    func : Function
        Function to decorate

    Returns
    -------
    Function : wrapped function
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        _detect_or_create_folders(args[0])
        return func(*args, **kwargs)

    return new_func


@detect_base_folder
def create_synthetic_genome(base_folder):
    """Create a synthetic genome with unique and multimapping regions annotated
    as in TEST_CHR_BED
    
    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created
        Genome will base saved as base_folder/fasta/chrA.fa
    
    Returns
    -------
    str : synthetic genome sequence
    """
    ustart = 100
    ulength = 1000
    spacer_length = 100

    #uend = ustart + ulength
    #dstart = uend + spacer_length
    dlength = 500
    dspacer = 50

    #dstart = uend + spacer_length
    #dend   = dstart + dlength + dspacer + dlength

    #splice_start = dend + spacer_length
    splice_flank_length = 25
    splice_spacer = 75

    # generate sequence
    unique_region = random_seq(ulength)
    duplicated_repeat = random_seq(dlength)
    duplicated_region = duplicated_repeat + ("N" * dspacer) + duplicated_repeat
    splice_region = random_seq(splice_flank_length) + ("N" * splice_spacer
                                                       ) + random_seq(splice_flank_length)

    # write to FASTA
    total_genome = (
        "N" * ustart
    ) + unique_region + "N" * spacer_length + duplicated_region + "N" * spacer_length + splice_region + "N" * 100
    with open(os.path.join(base_folder, _TEST_FILES["genome"]), "w") as fh:
        fh.write(">%s\n%s\n" % ("chrA", total_genome))
        fh.close()

    return total_genome


@detect_base_folder
def create_bowtie_index(
        base_folder, bowtie_location=os.path.join(os.path.sep, "usr", "local", "bin")
):
    """Build bowtie indices to enable alignments in bowtie and tophat
    against synthetic genome. Creates bowtie indices in base_folder/ebwt. 
    Requires a FASTA file of the synthetic genome in base_folder/fasta/chrA.fa,
    so run create_synthetic_genome() first

    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created
    
    bowtie_location : str
        path to folder containing bowtie-build
    
    Returns
    -------
    int : exit status of bowtie-build
    """
    unspliced_args = [
        os.path.join(bowtie_location, "bowtie-build"),
        os.path.join(base_folder, _TEST_FILES["genome"]),
        os.path.join(base_folder, _TEST_FILES["bowtie_index"])
    ]

    unspliced_exit = subprocess.call(unspliced_args)
    return unspliced_exit  #| spliced_exit


def _ndarray_to_variable_step(count_vec, fh, name):
    """Write a numpy.ndarray to a variableStep wiggle file
    
    Parameters
    ----------
    count_vec : numpy.ndarray
        vector of counts
    
    fh : file-like
        open filehandle
    
    name : str
        Track name
    """
    fh.write("track type=wiggle_0 name=%s\n" % name)
    fh.write("variableStep chrom=chrA span=1\n")
    for i in count_vec.nonzero()[0]:
        val = count_vec[i]
        fh.write("%s\t%s\n" % (i + 1, val))


def _ndarray_to_bedgraph(count_vec, fh, name):
    """Write a numpy.ndarray to a BEDGraph file
    
    Parameters
    ----------
    count_vec : numpy.ndarray
        vector of counts
    
    fh : file-like
        open filehandle
    
    name : str
        Track name
    """
    fh.write("track type=bedGraph name=%s\n" % name)
    last_val = count_vec[0]
    start_i = 0
    for i, val in enumerate(count_vec):
        if val != last_val:
            fh.write("%s\t%s\t%s\t%s\n" % ("chrA", start_i, i, last_val))
            start_i = i
            last_val = val
    fh.write("%s\t%s\t%s\t%s\n" % ("chrA", start_i, i + 1, last_val))


@detect_base_folder
def generate_reads(
        base_folder, reads_per_region=DEFAULT_READS_PER_REGION, read_length=DEFAULT_READ_LENGTH
):
    """Generates 30-nucleotide reads from a genome created by create_synthetic_genome,
    choosing from uniquely-mapping and multimapping regions annotated in 
    TEST_CHR_BED. 10000 reads are generated for each region type. Reads are
    returned in FASTA format. Also saves a numpy array of how many reads are expected
    to align to each nucleotide position in the synthetic genome, if reads are mapped
    at their 5' ends.

    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created.
        Reads will base saved as base_folder/fasta/chrA_reads.fa.
        Count vectors will base saved as various text files in
        base_folder/count_vectors    

    reads_per_region : int
        Number of reads to generate in each region

    read_length : int
        Length of reads to generate

    Returns
    -------
    dict : dictionary of numpy.ndarrays corresponding to expected number of counts
           at each genomic position, under various read alignment mapping rules 
    """
    with open(os.path.join(base_folder, "fasta", "chrA.fa")) as fh:
        genome = SeqIO.to_dict(SeqIO.parse(fh), "fasta")

    len_A = len(genome["chrA"])
    # TODO : align
    count_vectors = {
        "fiveprime_0_fw": numpy.zeros(len_A).astype(int),
        "fiveprime_15_fw": numpy.zeros(len_A).astype(int),
        "threeprime_0_fw": numpy.zeros(len_A).astype(int),
        "threeprime_15_fw": numpy.zeros(len_A).astype(int),
        "center_0_fw": numpy.zeros(len_A),
        "center_12_fw": numpy.zeros(len_A),
        "fiveprime_0_rc": numpy.zeros(len_A).astype(int),
        "fiveprime_15_rc": numpy.zeros(len_A).astype(int),
        "threeprime_0_rc": numpy.zeros(len_A).astype(int),
        "threeprime_15_rc": numpy.zeros(len_A).astype(int),
        "center_0_rc": numpy.zeros(len_A),
        "center_12_rc": numpy.zeros(len_A),
    } # yapf: disable

    with open(os.path.join(base_folder, _TEST_FILES["reads"]), "w") as read_fh:

        regions = filter(lambda x: "intron" not in x.get_name()\
                                   and "entire" not in x.get_name(),
                                   fetch_regions())

        for region in regions:
            strand_key = STRAND_KEYS[region.spanning_segment.strand]
            my_seq = region.get_sequence(genome)

            # choose 5' read locations
            read_locs = numpy.random.randint(
                0, high=len(my_seq) - read_length + 1, size=reads_per_region
            )

            # generate FASTA File
            # and save read positions to count vectors under various alignment mapping rules
            for n, loc in enumerate(read_locs):
                # write reads
                read_fh.write(
                    ">%s_%s\n%s\n" % (region.get_name(), n, my_seq[loc:loc + read_length])
                )
                _, position, _ = region.get_genomic_coordinate(loc)

                # populate 5' and 3' mapped count vectors
                for offset in (0, 15):
                    _, position, _ = region.get_genomic_coordinate(loc + offset)
                    count_vectors["fiveprime_%s_%s" % (offset, strand_key)][position] += 1
                    _, position, _ = region.get_genomic_coordinate(loc + read_length - offset - 1)
                    count_vectors["threeprime_%s_%s" % (offset, strand_key)][position] += 1

                # populate center-mapped count vectors
                read_positions = region.get_subchain(loc, loc + read_length).get_position_list()
                assert len(read_positions) == read_length
                for pos in read_positions:
                    count_vectors["center_0_%s" % strand_key][pos] += 1.0 / len(read_positions)
                assert len(read_positions[12:-12]) == read_length - 24
                for pos in read_positions[12:-12]:
                    count_vectors["center_12_%s" % strand_key][pos] += 1.0 / (
                        len(read_positions) - 24
                    )

        for k, v in count_vectors.items():
            numpy.savetxt(os.path.join(base_folder, "count_vectors", "%s.txt" % k), v)

    # export 5' mapped BEDGraph files
    with open(os.path.join(base_folder, _TEST_FILES["bedgraph_fw"]), "w") as bedgraph_fw:
        _ndarray_to_bedgraph(count_vectors["fiveprime_0_fw"], bedgraph_fw, base_folder)

    with open(os.path.join(base_folder, _TEST_FILES["bedgraph_rc"]), "w") as bedgraph_rc:
        _ndarray_to_bedgraph(count_vectors["fiveprime_0_rc"], bedgraph_rc, base_folder)

    # export 5' mapped variableStep wiggle files
    with open(os.path.join(base_folder, _TEST_FILES["variable_step_fw"]), "w") as vs_fw:
        _ndarray_to_variable_step(count_vectors["fiveprime_0_fw"], vs_fw, base_folder)

    with open(os.path.join(base_folder, _TEST_FILES["variable_step_rc"]), "w") as vs_rc:
        _ndarray_to_variable_step(count_vectors["fiveprime_0_rc"], vs_rc, base_folder)

    return count_vectors


@detect_base_folder
def perform_alignments(
        base_folder,
        bowtie_location=os.path.join(os.path.sep, "usr", "local", "bin"),
        tophat_location=os.path.join(os.path.sep, "usr", "local", "bin"),
        samtools_location=os.path.join(os.path.sep, "usr", "local", "bin"),
):
    """Perform alignments of generated reads against synthetic genome,
    in both Tophat and Bowtie so that both BAM and bowtie input may be tested.
    Note: spliced reads will not align in bowtie.
    
    Read alignments will be placed in base_folder/align
    
    Requires a bowtie index of the synthetic genome in base_folder/ebwt, and
    syntheti reads to align in base_folder/fasta/chrA_reads.fa 
    so run build_bowtie_index() and generate_reads() first.

    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created.
    
    bowtie_location : str
        path to folder containing bowtie executable

    tophat_location : str
        path to folder containing tophat executable

    samtools_location : str
        path to folder containing samtools executable
    
    Returns
    -------
    int : ORed exit status of bowtie and tophat (0 if both successful; 1 otherwise)
    """
    # align with no mismatches, choosing 1 random alignment from repeat regions
    bowtie_args = [
        os.path.join(bowtie_location, "bowtie"), "-v0", "-k1", "--best", "-f", "--un",
        os.path.join(base_folder, "align", "chrA_unspliced_unaligned.fa"),
        os.path.join(base_folder, _TEST_FILES["bowtie_index"]),
        os.path.join(base_folder, _TEST_FILES["reads"]),
        os.path.join(base_folder, _TEST_FILES["bowtie"])
    ]

    # align with no mismatches, choosing 1 random alignment from repeat regions
    tophat_args = [
        os.path.join(bowtie_location, "tophat"),
        "--bowtie1",
        "--read-mismatches=0",
        "--min-intron-length=20",
        "--library-type=fr-firststrand",
        "--raw-juncs",
        os.path.join(base_folder, _TEST_FILES["juncs"]),
        "--no-novel-juncs",
        "-o",
        os.path.join(base_folder, "align", "tophat"),
        os.path.join(base_folder, _TEST_FILES["bowtie_index"]),
        os.path.join(base_folder, _TEST_FILES["reads"]),
    ]

    samtools_multi_args = [
        os.path.join(samtools_location, "samtools"), "view", "-F", "256",
        os.path.join(base_folder, "align", "tophat", "accepted_hits.bam"), "-b", "-o",
        os.path.join(base_folder, "align", "chrA_tophat.bam")
    ]

    samtools_index_args = [
        os.path.join(samtools_location, "samtools"), "index",
        os.path.join(base_folder, "align", "chrA_tophat.bam")
    ]

    bowtie_exit = subprocess.call(bowtie_args)
    tophat_exit = subprocess.call(tophat_args)

    samtools_exit_1 = subprocess.call(samtools_multi_args)
    samtools_exit_2 = subprocess.call(samtools_index_args)

    return bowtie_exit | tophat_exit | samtools_exit_1 | samtools_exit_2


def create_dataset(
        base_folder,
        bowtie_location=os.path.join(os.path.sep, "usr", "local", "bin"),
        tophat_location=os.path.join(os.path.sep, "usr", "local", "bin"),
        samtools_location=os.path.join(os.path.sep, "usr", "local", "bin"),
):
    """Create a ground-truth dataset for testing |GenomeArray|
    
    This dataset includes a synthetic genome of random sequence, containing
    uniquely-mapping and multimapping regions; short sequence reads generated
    from these regions; alignments of these reads made in bowtie and tophat;
    and saved numpy tables indicating how many counts should appear at each
    genomic position under various mapping rules and offsets.
    
    Parameters
    ----------
    base_folder : str
        path to base folder in which test data will be created.
    
    bowtie_location : str
        path to folder containing bowtie executable

    tophat_location : str
        path to folder containing tophat executable

    samtools_location : str
        path to folder containing samtools executable
    
    Returns
    -------
    dict: dictionary containing the genome sequence, count vectors,
          and alignment status
    """
    dtmp = {}
    dtmp["base_folder"] = base_folder
    dtmp["genome_str"] = create_synthetic_genome(base_folder)
    dtmp["aligned"] = False
    if create_bowtie_index(base_folder, bowtie_location=bowtie_location) == 0:
        dtmp.update(generate_reads(base_folder))

        if perform_alignments(base_folder, bowtie_location=bowtie_location,
                              tophat_location=tophat_location,
                              samtools_location=samtools_location) == 0:
            dtmp["aligned"] = True
            return dtmp

    return dtmp



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.counts_in_region`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.util.services.decorators import catch_stderr
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    RPATH, REF_FILES,
    COUNT_OPTIONS,
    MASK_OPTIONS,
)
from plastid.bin.counts_in_region import main

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.counts_in_region",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="counts_in_region"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
_mask_outfile = os.path.join(test_info["temp_file_path"], "test_counts_in_region_mask.txt")
_no_mask_outfile = os.path.join(test_info["temp_file_path"], "test_counts_in_region_no_mask.txt")
_annotation_options = "--annotation_format BED --annotation_files %s" % REF_FILES["yeast_mini_bed"]
counts_in_region_tests = [
    # test with crossmap mask
    (
        "%s %s %s %s" % (_mask_outfile, COUNT_OPTIONS, MASK_OPTIONS, _annotation_options),
        [REF_FILES["yeast_counts_in_region_mask"]],
        [_mask_outfile], ["--sort_keys region_name"]
    ),
    # test without crossmap mask
    (
        "%s %s %s" % (_no_mask_outfile, COUNT_OPTIONS, _annotation_options),
        [REF_FILES["yeast_counts_in_region_no_mask"]],
        [_no_mask_outfile], ["--sort_keys region_name"]
    ),
    # test setting sum
    (
        "%s %s %s --sum 1e9" % (_no_mask_outfile, COUNT_OPTIONS, _annotation_options),
        [REF_FILES["yeast_counts_in_region_sum_1"]],
        [_no_mask_outfile], ["--sort_keys region_name"]
    ),
]
"""Functional tests of :py:mod:`plastid.bin.counts_in_region`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for plastid.bin.counts_in_region"""
    for x in execute_helper(test_info, counts_in_region_tests):
        yield x



#!/usr/bin/env python
"""Utility functions for systematically naming splice junctions and retrieving their sequences
"""
import re

#===============================================================================
# INDEX: Utility functions for naming splice junctions and retrieving sequences
#===============================================================================

junction_pat = re.compile(r"([^@]+)@([0-9]+)\^([0-9]+)\(([+-])\)f([0-9]+)")
"""Regex for matching splice junctions in plastid's notation"""


def get_junction_tuple(ivc):
    """Convert a |SegmentChain| representing a splice junction to a tuple
    
    Parameters
    ----------
    ivc : |SegmentChain|
         A two-exon fragment representing a splice junction
        
    Returns
    -------
    tuple
        `(chromosome name, half-open end of fiveprime exon, first position of threeprime exon, strand)`
    """
    return (ivc.spanning_segment.chrom, ivc[0].end, ivc[1].start, ivc.spanning_segment.strand)



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.get_count_vectors`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    RPATH,
    REF_FILES,
    COUNT_OPTIONS,
    ANNOTATION_OPTIONS,
    MASK_OPTIONS,
)
from plastid.bin.get_count_vectors import main
from plastid.util.services.decorators import catch_stderr
from plastid.util.services.mini2to3 import cStringIO

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

get_count_vectors_test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.get_count_vectors",
    "ref_file_path": resource_filename("plastid", "test/data/command_line/gen_count_vectors"),
    "temp_file_path": tempfile.mkdtemp(prefix="get_count_vectors"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
ref_files = []
out_files = []
with open(REF_FILES["yeast_mini_bed"]) as fh:
    for line in fh:
        fn = line.split("\t")[3]
        ref_files.append(os.path.join(get_count_vectors_test_info["ref_file_path"], "%s.txt" % fn))
        out_files.append(os.path.join(get_count_vectors_test_info["temp_file_path"], "%s.txt" % fn))

get_count_vectors_tests = [
    (
        get_count_vectors_test_info["temp_file_path"] + COUNT_OPTIONS + MASK_OPTIONS +
        " --annotation_format BED --annotation_file " + REF_FILES["yeast_mini_bed"], ref_files,
        out_files, ["--no_header"] * len(ref_files)
    )
]
"""Functional tests of :py:mod:`plastid.bin.get_count_vectors`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""


@attr(test="functional")
@attr(speed="slow")
def do_test():
    for x in execute_helper(get_count_vectors_test_info, get_count_vectors_tests):
        yield x



#!/usr/bin/env python
import sys
import os
import fcntl
import types
import warnings
from nose.plugins.attrib import attr
from nose.tools import assert_equal, assert_true, assert_greater_equal, assert_raises, assert_set_equal, assert_not_equal
from plastid.util.services.decorators import (
    notimplemented,
    deprecated,
    parallelize,
    in_separate_process,
    catch_stderr,
    catch_stdout,
    catch_warnings,
)
#===============================================================================
# INDEX: functions and classes that will be decorated
#===============================================================================


def stderr_func(msg):
    sys.stderr.write(msg)
    return msg


def stdout_func(msg):
    sys.stdout.write(msg)
    return msg


def util_func(x):
    """Square numbers and return process in which function was run
    
    Parameters
    ----------
    x : int or float

    Returns
    -------
    int or float
        Squared value of ``x``
    
    int
        Process ID in which function was run
    """
    return x**2, os.getpid()


def func_that_warns():
    warnings.warn("Some warning", UserWarning)


def get_pipes():
    readfd, writefd = os.pipe()
    fcntl.fcntl(readfd, fcntl.F_SETFL, os.O_NONBLOCK)
    fcntl.fcntl(writefd, fcntl.F_SETFL, os.O_NONBLOCK)
    return os.fdopen(readfd, "r"), os.fdopen(writefd, "w")


class UtilClass(object):
    def __init__(self, tmp):
        self.name = tmp

    def get_name(self):
        return self.name


#===============================================================================
# INDEX: unit tests
#===============================================================================

# stdout/err redirection -------------------------------------------------------
STDERR_FD = sys.stderr.fileno()

if sys.version_info[0] == 2:
    EMPTY_BUFFER_ERROR = IOError
else:
    # Python 3.x codecs returns None from empty non-blocking buffers
    # which causes a TypeError to be raised instead of
    # the IOError raised in Python 2.x
    EMPTY_BUFFER_ERROR = TypeError


@attr(test="unit")
def test_catch_stderr_doesnt_print_without_buffer():
    # spy on `inner` by making sure there is nothing written to stderr
    outer_reader, outer_writer = get_pipes()
    message = "this is a test"
    sys.stderr = sys.__stderr__

    @catch_stderr(outer_writer)
    def inner():
        # make sure value is returned from wrapped function
        wrapped = catch_stderr()(stderr_func)
        msg = wrapped(message)
        assert_equal(msg, message)

    inner()
    outer_writer.flush()
    outer_writer.close()

    # make sure no message made it out of `inner`
    # with nothing to read in pipe, outer_reader should raise IOError
    assert_raises(EMPTY_BUFFER_ERROR, outer_reader.read)
    outer_reader.close()


@attr(test="unit")
def test_catch_stderr_doesnt_print_with_buffer_but_catches_in_buffer():
    # spy on `inner` by making sure there is nothing written to stderr
    # but make sure message is found in inner readre
    outer_reader, outer_writer = get_pipes()
    message = "this is a test"
    sys.stderr = sys.__stderr__

    @catch_stderr(outer_writer)
    def inner():
        inner_reader, inner_writer = get_pipes()
        wrapped = catch_stderr(inner_writer)(stderr_func)

        # make sure value is returned from wrapped function
        msg = wrapped(message)
        assert_equal(message, msg)

        # make sure we caught entire message from stderr
        inner_writer.flush()
        inner_writer.close()
        assert_equal(message, inner_reader.read())
        inner_reader.close()

    inner()
    outer_writer.flush()
    outer_writer.close()

    # make sure no message made it out of `inner`
    # with nothing to read in pipe, outer_reader should raise IOError
    assert_raises(EMPTY_BUFFER_ERROR, outer_reader.read)
    outer_reader.close()


# catch warnings    ------------------------------------------------------------


@attr(test="unit")
def test_catch_warnings_catches_warnings():
    ign_func = catch_warnings("ignore")(func_that_warns)

    num = 5
    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        for x in range(num):
            func_that_warns()

    # make sure warning is issued with vanilla function
    assert_equal(len(warns), num)

    num = 5
    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        for x in range(num):
            ign_func()

    # make sure warning is caught with wrapped function
    assert_equal(len(warns), 0)


# notimplemented    ------------------------------------------------------------


@attr(test="unit")
def test_notimplemented_raises_exception():
    my_func = notimplemented(util_func)
    assert_true(isinstance(my_func, types.FunctionType))

    assert_raises(NotImplementedError, my_func, 5)


# deprecated    ----------------------------------------------------------------


@attr(test="unit")
def test_deprecated_function_raises_warning_only_once():
    num = 5
    my_func = deprecated(util_func)
    assert_true(isinstance(my_func, types.FunctionType))

    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        for x in range(num):
            assert_equal(my_func(x), util_func(x))
            my_func(x)

    # make sure warning is issued only once
    assert_equal(len(warns), 1)


@attr(test="unit")
def test_deprecated_class_raises_warning():
    reg_obj = UtilClass("my_object")
    dep_class = deprecated(UtilClass)

    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        dep_obj = dep_class("my_object")
        assert_true(isinstance(dep_obj, UtilClass))

    # make sure warning is issued
    assert_equal(len(warns), 1)

    # make sure wrapped class behaves as it should
    assert_equal(reg_obj.get_name(), dep_obj.get_name())


# parallelize or in other processes --------------------------------------------


@attr(test="unit")
def test_parallelize_spawns_processes_and_gets_correct_ansswer():
    x = range(500)
    my_func = parallelize(util_func)
    outer_vals, outer_pids = zip(*[util_func(X) for X in x])
    inner_vals, inner_pids = zip(*my_func(x))

    assert_set_equal(set(outer_vals), set(inner_vals))
    assert_not_equal(set(outer_pids), set(inner_pids))


@attr(test="unit")
def test_in_separate_process_spawns_process_and_gets_correct_ansswers():
    my_func = in_separate_process(util_func)
    assert_true(isinstance(my_func, types.FunctionType))

    for x in range(100):
        res_out, outer_pid = my_func(x)
        res_in, inner_pid = util_func(x)
        assert_equal(res_out, res_in)
        assert_not_equal(outer_pid, inner_pid)
